from threading import Lock

from customer_services.model import UpdateCustomerException
from typing import Any

from msgbus import MBEasyContext
from mwhelper import BaseRepository
from persistence import Connection

from customer_services.model import InsertCustomerException, Address, Customer
from customer_services.model.repository import CustomerRepository as CustomerServicesRepository


class CustomerRepository(CustomerServicesRepository, BaseRepository):

    def __init__(self, mb_context, persistence_name, database_lock):
        # type: (MBEasyContext, str, Lock) -> None

        super(CustomerRepository, self).__init__(mb_context)

        self.persistence_name = persistence_name
        self.database_lock = database_lock

    def get_customer(self, phone):
        # type: (str) -> Any[Customer, None]

        try:
            def inner_func(conn):
                # type: (Connection) -> Any[Customer, None]

                query = """
                            SELECT
                            c.Phone,
                            c.Document,
                            c.Name,
                            ca.Id,
                            ca.ZipCode,
                            ca.State,
                            ca.City,
                            ca.Street,
                            ca.NeighborHood,
                            ca.Number,
                            ca.Complement,
                            ca.ReferencePoint
                            FROM Customer c
                            JOIN CustomerAddress ca ON ca.Id = c.AddressId
                            WHERE c.Phone = '{}'
                        """
                query_format = query.format(phone)
                cursor = conn.select(query_format)
                if not cursor.rows():
                    return None

                row = cursor.get_row(0)
                address = Address(
                    row.get_entry("Id"),
                    row.get_entry("ZipCode"),
                    row.get_entry("State"),
                    row.get_entry("City"),
                    row.get_entry("Street"),
                    row.get_entry("Neighborhood"),
                    row.get_entry("Number"),
                    row.get_entry("Complement"),
                    row.get_entry("ReferencePoint")
                )
                return Customer(row.get_entry("Phone"), row.get_entry("Document"), row.get_entry("Name"), address)

            return self.execute_with_connection(inner_func, service_name=self.persistence_name)
        except Exception as _:
            raise InsertCustomerException()

    def add_customer(self, customer):
        # type: (Customer) -> None

        with self.database_lock:
            try:
                def inner_func(conn):
                    # type: (Connection) -> None

                    self._insert_address(customer.address)
                    address_id = self._get_last_inserted_address()

                    query = """
                                INSERT INTO Customer(Phone, Document, Name, AddressId)
                                VALUES ('{}', '{}', '{}', '{}')
                            """
                    query_format = query.format(customer.phone, customer.document, customer.name, address_id)
                    conn.query(query_format)

                self.execute_with_connection(inner_func, service_name=self.persistence_name)
            except Exception as _:
                raise InsertCustomerException()

    def update_customer(self, customer):
        # type: (Customer) -> None

        with self.database_lock:
            try:
                def inner_func(conn):
                    # type: (Connection) -> None

                    self._update_address(customer.address)
                    query = """
                                UPDATE Customer
                                SET
                                Document='{}',
                                Name='{}'
                                WHERE Phone='{}'
                            """
                    query_format = query.format(customer.document, customer.name, customer.phone)
                    conn.query(query_format)

                self.execute_with_connection(inner_func, service_name=self.persistence_name)
            except Exception as _:
                raise InsertCustomerException()

    def _insert_address(self, address):
        # type: (Address) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                query = """
                            INSERT INTO CustomerAddress(
                                ZipCode, 
                                State, 
                                City, 
                                Street, 
                                Neighborhood, 
                                Number, 
                                Complement, 
                                ReferencePoint
                            )
                            VALUES ('{}', '{}', '{}', '{}', '{}', '{}', '{}', '{}')
                        """
                query_format = query.format(
                    address.zip_code,
                    address.state,
                    address.city,
                    address.street,
                    address.neighborhood,
                    address.number,
                    address.complement,
                    address.reference_point
                )
                conn.query(query_format)

            self.execute_with_connection(inner_func, service_name=self.persistence_name)
        except Exception as _:
            raise UpdateCustomerException()

    def _get_last_inserted_address(self):
        # type: () -> int

        try:
            def inner_func(conn):
                # type: (Connection) -> int

                query = """
                            SELECT Id FROM CustomerAddress
                            ORDER BY Id DESC
                            LIMIT 1
                        """
                cursor = conn.select(query)
                if not cursor.rows():
                    return -1

                row = cursor.get_row(0)
                return row.get_entry("Id")

            return self.execute_with_connection(inner_func, service_name=self.persistence_name)
        except Exception as _:
            raise InsertCustomerException()

    def _update_address(self, address):
        # type: (Address) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                query = """
                            UPDATE CustomerAddress
                            SET
                            ZipCode='{}',
                            State='{}',
                            City='{}',
                            Street='{}',
                            Neighborhood='{}',
                            Number='{}',
                            Complement='{}',
                            ReferencePoint='{}'
                            WHERE Id='{}'
                        """
                query_format = query.format(
                    address.zip_code,
                    address.state,
                    address.city,
                    address.street,
                    address.neighborhood,
                    address.number,
                    address.complement,
                    address.reference_point,
                    address.address_id
                )
                conn.query(query_format)

            self.execute_with_connection(inner_func, service_name=self.persistence_name)
        except Exception as _:
            raise InsertCustomerException()



from logging import Logger

from typing import Optional


class Configurations(object):
    def __init__(
        self,
        required_services=None,         # type: Optional[str]
        service_name=None,              # type: Optional[str]
        persistence_name=None,          # type: Optional[str]
        store_id=None,                  # type: Optional[int]
        logger=None,                    # type: Optional[Logger]
        base_url=None,                  # type: Optional[str]
        api_key=None,                   # type: Optional[str]
        id_type=None,                   # type: Optional[str]
        is_loyalty_enabled=True,        # type: Optional[bool]
        partner_retry_policy=None,      # type: Optional[RetryPolicy]
    ):
        # type: (...) -> None

        self.required_services = required_services
        self.service_name = service_name
        self.persistence_name = persistence_name
        self.store_id = store_id
        self.logger = logger
        self.base_url = base_url
        self.api_key = api_key
        self.id_type = id_type
        self.is_loyalty_enabled = is_loyalty_enabled
        self.endpoints = Endpoints()
        if partner_retry_policy is None:
            partner_retry_policy = RetryPolicy(
                count=1,
                sleep=1,
            )
        self.partner_retry_policy = partner_retry_policy


class Endpoints(object):
    def __init__(self):
        self.retrieve_loyalty_id = None
        self.get_benefit = None
        self.burn_benefit = None
        self.unlock_benefit = None


class RetryPolicy(object):
    def __init__(
        self,
        count,   # type: int
        sleep,   # type: int
    ):
        self.count = count
        self.sleep = sleep



from production.box._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import ProdStates


class OrderMergerBox(OrderChangerProductionBox):
    def __init__(self, name, sons, logger=None):
        super(OrderMergerBox, self).__init__(name, sons, logger)

    def change_order(self, order):
        if hasattr(order, "stamped") and order.prod_state == ProdStates.INVALID:
            return None

        stored_order = self.get_order(order.order_id)
        if stored_order and hasattr(order, "stamped"):
            order.items = self._get_stamped_items(stored_order, order)
            order.prod_state = self._get_new_order_prod_state(order, stored_order)

        if order.purged:
            self.delete_order(order.order_id)
        else:
            self.save_order(order)
        return order

    @staticmethod
    def _get_new_order_prod_state(order, stored_order):
        all_states = [x.prod_state for x in [order, stored_order]]
        state_values = {ProdStates.NORMAL: 3,
                        ProdStates.SERVED: 2,
                        ProdStates.INVALID: 1}
        state_weights = []
        for state in all_states:
            if state in state_values:
                state_weights.append(state_values[state])

        if not state_weights:
            return order.prod_state

        return next((state for state, weight in state_values.items() if weight == max(state_weights)))

    def _get_stamped_items(self, stored_order, order):
        stamped_items = stored_order.items
        stamped_order_item_codes = self._list_order_item_codes(stored_order)
        for item in order.items:
            if hasattr(item, "stamped") and item.item_code not in stamped_order_item_codes:
                stamped_items.append(item)
        return stamped_items

    @staticmethod
    def _list_order_item_codes(order):
        item_ids = {}
        for item in order.items:
            item_ids[item.item_code] = ""
        return item_ids



#  -*- coding: utf-8 -*-
""" ESC/POS Exceptions classes

Result/Exit codes:

    - `0`  = success
    - `10` = No Barcode type defined :py:exc:`~escpos.exceptions.BarcodeTypeError`
    - `20` = Barcode size values are out of range :py:exc:`~escpos.exceptions.BarcodeSizeError`
    - `30` = Barcode text not supplied :py:exc:`~escpos.exceptions.BarcodeCodeError`
    - `40` = Image height is too large :py:exc:`~escpos.exceptions.ImageSizeError`
    - `50` = No string supplied to be printed :py:exc:`~escpos.exceptions.TextError`
    - `60` = Invalid pin to send Cash Drawer pulse :py:exc:`~escpos.exceptions.CashDrawerError`
    - `70` = Invalid number of tab positions :py:exc:`~escpos.exceptions.TabPosError`
    - `80` = Invalid char code :py:exc:`~escpos.exceptions.CharCodeError`
    - `90` = USB device not found :py:exc:`~escpos.exceptions.USBNotFoundError`
    - `100` = Set variable out of range :py:exc:`~escpos.exceptions.SetVariableError`
    - `200` = Configuration not found :py:exc:`~escpos.exceptions.ConfigNotFoundError`
    - `210` = Configuration syntax error :py:exc:`~escpos.exceptions.ConfigSyntaxError`
    - `220` = Configuration section not found :py:exc:`~escpos.exceptions.ConfigSectionMissingError`

:author: `Manuel F Martinez <manpaz@bashlinux.com>`_ and others
:organization: Bashlinux and `python-escpos <https://github.com/python-escpos>`_
:copyright: Copyright (c) 2012 Bashlinux
:license: GNU GPL v3
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class Error(Exception):
    """ Base class for ESC/POS errors """
    def __init__(self, msg, status=None):
        Exception.__init__(self)
        self.msg = msg
        self.resultcode = 1
        if status is not None:
            self.resultcode = status

    def __str__(self):
        return self.msg


class BarcodeTypeError(Error):
    """ No Barcode type defined.

    This exception indicates that no known barcode-type has been entered. The barcode-type has to be
    one of those specified in :py:meth:`escpos.escpos.Escpos.barcode`.
    The returned error code is `10`.
    """
    def __init__(self, msg=""):
        Error.__init__(self, msg)
        self.msg = msg
        self.resultcode = 10

    def __str__(self):
        return "No Barcode type is defined ({msg})".format(msg=self.msg)


class BarcodeSizeError(Error):
    """ Barcode size is out of range.

    This exception indicates that the values for the barcode size are out of range.
    The size of the barcode has to be in the range that is specified in :py:meth:`escpos.escpos.Escpos.barcode`.
    The resulting returncode is `20`.
    """
    def __init__(self, msg=""):
        Error.__init__(self, msg)
        self.msg = msg
        self.resultcode = 20

    def __str__(self):
        return "Barcode size is out of range ({msg})".format(msg=self.msg)


class BarcodeCodeError(Error):
    """ No Barcode code was supplied.

    No data for the barcode has been supplied in :py:meth:`escpos.escpos.Escpos.barcode`.
    The returncode for this exception is `30`.
    """
    def __init__(self, msg=""):
        Error.__init__(self, msg)
        self.msg = msg
        self.resultcode = 30

    def __str__(self):
        return "No Barcode code was supplied"


class ImageSizeError(Error):
    """ Image height is longer than 255px and can't be printed.

    The returncode for this exception is `40`.
    """
    def __init__(self, msg=""):
        Error.__init__(self, msg)
        self.msg = msg
        self.resultcode = 40

    def __str__(self):
        return "Image height is longer than 255px and can't be printed"


class TextError(Error):
    """ Text string must be supplied to the `text()` method.

    This exception is raised when an empty string is passed to :py:meth:`escpos.escpos.Escpos.text`.
    The returncode for this exception is `50`.
    """
    def __init__(self, msg=""):
        Error.__init__(self, msg)
        self.msg = msg
        self.resultcode = 50

    def __str__(self):
        return "Text string must be supplied to the text() method"


class CashDrawerError(Error):
    """ Valid pin must be set in order to send pulse.

    A valid pin number has to be passed onto the method :py:meth:`escpos.escpos.Escpos.cashdraw`.
    The returncode for this exception is `60`.
    """
    def __init__(self, msg=""):
        Error.__init__(self, msg)
        self.msg = msg
        self.resultcode = 60

    def __str__(self):
        return "Valid pin must be set to send pulse"


class TabPosError(Error):
    """ Valid tab positions must be in the range 0 to 16.

    This exception is raised by :py:meth:`escpos.escpos.Escpos.control`.
    The returncode for this exception is `70`.
    """
    def __init__(self, msg=""):
        Error.__init__(self, msg)
        self.msg = msg
        self.resultcode = 70

    def __str__(self):
        return "Valid tab positions must be in the range 0 to 16"


class CharCodeError(Error):
    """ Valid char code must be set.

    The supplied charcode-name in :py:meth:`escpos.escpos.Escpos.charcode` is unknown.
    Ths returncode for this exception is `80`.
    """
    def __init__(self, msg=""):
        Error.__init__(self, msg)
        self.msg = msg
        self.resultcode = 80

    def __str__(self):
        return "Valid char code must be set"


class USBNotFoundError(Error):
    """ Device wasn't found (probably not plugged in)

    The USB device seems to be not plugged in.
    Ths returncode for this exception is `90`.
    """
    def __init__(self, msg=""):
        Error.__init__(self, msg)
        self.msg = msg
        self.resultcode = 90

    def __str__(self):
        return "USB device not found"


class SetVariableError(Error):
    """ A set method variable was out of range

    Check set variables against minimum and maximum values
    Ths returncode for this exception is `100`.
    """
    def __init__(self, msg=""):
        Error.__init__(self, msg)
        self.msg = msg
        self.resultcode = 100

    def __str__(self):
        return "Set variable out of range"


# Configuration errors

class ConfigNotFoundError(Error):
    """ The configuration file was not found

    The default or passed configuration file could not be read
    Ths returncode for this exception is `200`.
    """
    def __init__(self, msg=""):
        Error.__init__(self, msg)
        self.msg = msg
        self.resultcode = 200

    def __str__(self):
        return "Configuration not found ({msg})".format(msg=self.msg)


class ConfigSyntaxError(Error):
    """ The configuration file is invalid

    The syntax is incorrect
    Ths returncode for this exception is `210`.
    """
    def __init__(self, msg=""):
        Error.__init__(self, msg)
        self.msg = msg
        self.resultcode = 210

    def __str__(self):
        return "Configuration syntax is invalid ({msg})".format(msg=self.msg)


class ConfigSectionMissingError(Error):
    """ The configuration file is missing a section

    The part of the config asked for doesn't exist in the loaded configuration
    Ths returncode for this exception is `220`.
    """
    def __init__(self, msg=""):
        Error.__init__(self, msg)
        self.msg = msg
        self.resultcode = 220

    def __str__(self):
        return "Configuration section is missing ({msg})".format(msg=self.msg)



from production.box import ReprocessOrdersBox
from production.model import ReprocessTimeNotFoundException
from production.treebuilder import BoxConfiguration
from production.treebuilder.boxfactory.builder._BoxBuilder import BoxBuilder


class ReprocessOrdersBoxBuilder(BoxBuilder):
    def __init__(self, publish_scheduler, loggers):
        super(ReprocessOrdersBoxBuilder, self).__init__(loggers)
        self.publish_scheduler = publish_scheduler
        self.loggers = loggers

    def build(self, config):
        # type: (BoxConfiguration) -> ReprocessOrdersBox

        reprocess_time = self.get_extra(config, "ReprocessTimeMinutes", None)
        reprocess_display_time = self.get_extra(config, "ReprocessDisplayTimeMinutes", None)

        if reprocess_time is None and reprocess_display_time is None:
            raise ReprocessTimeNotFoundException("ReprocessOrdersBoxBuilder without any time in configuration")

        reprocess_time_secs = int(reprocess_time) * 60 if reprocess_time is not None else None
        reprocess_display_time_sec = int(reprocess_display_time) * 60 if reprocess_display_time is not None else None
        logger = self.get_logger(config)

        return ReprocessOrdersBox(
            config.name,
            config.sons,
            self.publish_scheduler,
            reprocess_time_secs,
            reprocess_display_time_sec,
            logger
        )



data = (
'bbweok',    # 0x00
'bbweot',    # 0x01
'bbweop',    # 0x02
'bbweoh',    # 0x03
'bbwe',    # 0x04
'bbweg',    # 0x05
'bbwegg',    # 0x06
'bbwegs',    # 0x07
'bbwen',    # 0x08
'bbwenj',    # 0x09
'bbwenh',    # 0x0a
'bbwed',    # 0x0b
'bbwel',    # 0x0c
'bbwelg',    # 0x0d
'bbwelm',    # 0x0e
'bbwelb',    # 0x0f
'bbwels',    # 0x10
'bbwelt',    # 0x11
'bbwelp',    # 0x12
'bbwelh',    # 0x13
'bbwem',    # 0x14
'bbweb',    # 0x15
'bbwebs',    # 0x16
'bbwes',    # 0x17
'bbwess',    # 0x18
'bbweng',    # 0x19
'bbwej',    # 0x1a
'bbwec',    # 0x1b
'bbwek',    # 0x1c
'bbwet',    # 0x1d
'bbwep',    # 0x1e
'bbweh',    # 0x1f
'bbwi',    # 0x20
'bbwig',    # 0x21
'bbwigg',    # 0x22
'bbwigs',    # 0x23
'bbwin',    # 0x24
'bbwinj',    # 0x25
'bbwinh',    # 0x26
'bbwid',    # 0x27
'bbwil',    # 0x28
'bbwilg',    # 0x29
'bbwilm',    # 0x2a
'bbwilb',    # 0x2b
'bbwils',    # 0x2c
'bbwilt',    # 0x2d
'bbwilp',    # 0x2e
'bbwilh',    # 0x2f
'bbwim',    # 0x30
'bbwib',    # 0x31
'bbwibs',    # 0x32
'bbwis',    # 0x33
'bbwiss',    # 0x34
'bbwing',    # 0x35
'bbwij',    # 0x36
'bbwic',    # 0x37
'bbwik',    # 0x38
'bbwit',    # 0x39
'bbwip',    # 0x3a
'bbwih',    # 0x3b
'bbyu',    # 0x3c
'bbyug',    # 0x3d
'bbyugg',    # 0x3e
'bbyugs',    # 0x3f
'bbyun',    # 0x40
'bbyunj',    # 0x41
'bbyunh',    # 0x42
'bbyud',    # 0x43
'bbyul',    # 0x44
'bbyulg',    # 0x45
'bbyulm',    # 0x46
'bbyulb',    # 0x47
'bbyuls',    # 0x48
'bbyult',    # 0x49
'bbyulp',    # 0x4a
'bbyulh',    # 0x4b
'bbyum',    # 0x4c
'bbyub',    # 0x4d
'bbyubs',    # 0x4e
'bbyus',    # 0x4f
'bbyuss',    # 0x50
'bbyung',    # 0x51
'bbyuj',    # 0x52
'bbyuc',    # 0x53
'bbyuk',    # 0x54
'bbyut',    # 0x55
'bbyup',    # 0x56
'bbyuh',    # 0x57
'bbeu',    # 0x58
'bbeug',    # 0x59
'bbeugg',    # 0x5a
'bbeugs',    # 0x5b
'bbeun',    # 0x5c
'bbeunj',    # 0x5d
'bbeunh',    # 0x5e
'bbeud',    # 0x5f
'bbeul',    # 0x60
'bbeulg',    # 0x61
'bbeulm',    # 0x62
'bbeulb',    # 0x63
'bbeuls',    # 0x64
'bbeult',    # 0x65
'bbeulp',    # 0x66
'bbeulh',    # 0x67
'bbeum',    # 0x68
'bbeub',    # 0x69
'bbeubs',    # 0x6a
'bbeus',    # 0x6b
'bbeuss',    # 0x6c
'bbeung',    # 0x6d
'bbeuj',    # 0x6e
'bbeuc',    # 0x6f
'bbeuk',    # 0x70
'bbeut',    # 0x71
'bbeup',    # 0x72
'bbeuh',    # 0x73
'bbyi',    # 0x74
'bbyig',    # 0x75
'bbyigg',    # 0x76
'bbyigs',    # 0x77
'bbyin',    # 0x78
'bbyinj',    # 0x79
'bbyinh',    # 0x7a
'bbyid',    # 0x7b
'bbyil',    # 0x7c
'bbyilg',    # 0x7d
'bbyilm',    # 0x7e
'bbyilb',    # 0x7f
'bbyils',    # 0x80
'bbyilt',    # 0x81
'bbyilp',    # 0x82
'bbyilh',    # 0x83
'bbyim',    # 0x84
'bbyib',    # 0x85
'bbyibs',    # 0x86
'bbyis',    # 0x87
'bbyiss',    # 0x88
'bbying',    # 0x89
'bbyij',    # 0x8a
'bbyic',    # 0x8b
'bbyik',    # 0x8c
'bbyit',    # 0x8d
'bbyip',    # 0x8e
'bbyih',    # 0x8f
'bbi',    # 0x90
'bbig',    # 0x91
'bbigg',    # 0x92
'bbigs',    # 0x93
'bbin',    # 0x94
'bbinj',    # 0x95
'bbinh',    # 0x96
'bbid',    # 0x97
'bbil',    # 0x98
'bbilg',    # 0x99
'bbilm',    # 0x9a
'bbilb',    # 0x9b
'bbils',    # 0x9c
'bbilt',    # 0x9d
'bbilp',    # 0x9e
'bbilh',    # 0x9f
'bbim',    # 0xa0
'bbib',    # 0xa1
'bbibs',    # 0xa2
'bbis',    # 0xa3
'bbiss',    # 0xa4
'bbing',    # 0xa5
'bbij',    # 0xa6
'bbic',    # 0xa7
'bbik',    # 0xa8
'bbit',    # 0xa9
'bbip',    # 0xaa
'bbih',    # 0xab
'sa',    # 0xac
'sag',    # 0xad
'sagg',    # 0xae
'sags',    # 0xaf
'san',    # 0xb0
'sanj',    # 0xb1
'sanh',    # 0xb2
'sad',    # 0xb3
'sal',    # 0xb4
'salg',    # 0xb5
'salm',    # 0xb6
'salb',    # 0xb7
'sals',    # 0xb8
'salt',    # 0xb9
'salp',    # 0xba
'salh',    # 0xbb
'sam',    # 0xbc
'sab',    # 0xbd
'sabs',    # 0xbe
'sas',    # 0xbf
'sass',    # 0xc0
'sang',    # 0xc1
'saj',    # 0xc2
'sac',    # 0xc3
'sak',    # 0xc4
'sat',    # 0xc5
'sap',    # 0xc6
'sah',    # 0xc7
'sae',    # 0xc8
'saeg',    # 0xc9
'saegg',    # 0xca
'saegs',    # 0xcb
'saen',    # 0xcc
'saenj',    # 0xcd
'saenh',    # 0xce
'saed',    # 0xcf
'sael',    # 0xd0
'saelg',    # 0xd1
'saelm',    # 0xd2
'saelb',    # 0xd3
'saels',    # 0xd4
'saelt',    # 0xd5
'saelp',    # 0xd6
'saelh',    # 0xd7
'saem',    # 0xd8
'saeb',    # 0xd9
'saebs',    # 0xda
'saes',    # 0xdb
'saess',    # 0xdc
'saeng',    # 0xdd
'saej',    # 0xde
'saec',    # 0xdf
'saek',    # 0xe0
'saet',    # 0xe1
'saep',    # 0xe2
'saeh',    # 0xe3
'sya',    # 0xe4
'syag',    # 0xe5
'syagg',    # 0xe6
'syags',    # 0xe7
'syan',    # 0xe8
'syanj',    # 0xe9
'syanh',    # 0xea
'syad',    # 0xeb
'syal',    # 0xec
'syalg',    # 0xed
'syalm',    # 0xee
'syalb',    # 0xef
'syals',    # 0xf0
'syalt',    # 0xf1
'syalp',    # 0xf2
'syalh',    # 0xf3
'syam',    # 0xf4
'syab',    # 0xf5
'syabs',    # 0xf6
'syas',    # 0xf7
'syass',    # 0xf8
'syang',    # 0xf9
'syaj',    # 0xfa
'syac',    # 0xfb
'syak',    # 0xfc
'syat',    # 0xfd
'syap',    # 0xfe
'syah',    # 0xff
)



import json
import logging

from typing import (
    Any,
    Dict,
    List,
    Union,
)

from application.domain import TableService


class SectorOperatorInteractor(object):
    def __init__(
        self,
        table_service,  # type: TableService
    ):
        # type: (...) -> None

        self.logger = logging.getLogger(name="PosActions")
        self.table_service = table_service

    def get_sectors(
        self,
        pos_id,     # type: Any[Union[str, int]]
    ):
        # type: (...) -> List[str]

        return self.table_service.get_sectors(pos_id=pos_id)

    def get_operator_sector_info(
        self,
        pos_id,     # type: Any[Union[str, int]]
    ):
        # type: (...) -> Dict[str, List[int]]

        return self.table_service.get_operator_sector_info(pos_id=pos_id)

    def update_operator_sector_info(
        self,
        pos_id,         # type: Any[Union[str, int]]
        update_info,    # type: Dict[str, Dict]
    ):
        # type: (...) -> bool

        json_info = json.dumps(update_info) if update_info else "{}"
        return self.table_service.update_operator_sector_info(
            pos_id=pos_id,
            update_info=json_info,
        )



from pos_model import TenderType


class FinalizeActionFunction(object):
    confirm = "0"
    cancel = "1"


class CancelCodes(object):
    credit = 210
    debit = 211
    digital_wallet = 123


class PaymentCodes(object):
    credit = 3
    debit = 2
    digital_wallet = 122
    connection_test = 111


class ResultCodes(object):
    success = 0
    continue_loop = 10000


class CommandCodes(object):

    show_message_operator = 1
    show_message_customer = 2
    show_message_both = 3
    clear_message_operator = 11
    clear_message_customer = 12
    clear_message_both = 13
    idle_command = 23
    idle_field = 30
    set_amount_transaction = 34


class FieldCodes(object):
    empty_information = -1
    receipt_customer = 121
    receipt_merchant = 122
    nsu = 133
    nsu_host = 134
    nsu_origin = 135
    cancel_card_number = 512
    cancel_transaction_date = 515
    cancel_document_number = 516
    transaction_value = 146


class SitefFunctions(object):

    @staticmethod
    def get_cancel_code(
            tender_type  # type: int
    ):
        if int(tender_type) == TenderType.credit:
            return CancelCodes.credit

        if int(tender_type) == TenderType.debit:
            return CancelCodes.debit

        if int(tender_type) == TenderType.digital_wallet:
            return CancelCodes.digital_wallet

        return CancelCodes.credit

    @staticmethod
    def get_payment_code(
        electronic_type,    # type: int
    ):
        # type: (...) -> int

        if int(electronic_type) in (TenderType.credit, TenderType.gift_card):
            return PaymentCodes.credit

        if int(electronic_type) in (TenderType.debit, TenderType.voucher):
            return PaymentCodes.debit

        if int(electronic_type) == TenderType.digital_wallet:
            return PaymentCodes.digital_wallet

        return PaymentCodes.connection_test



import base64
import json
import os
import re
import time
from ctypes import (
    Array,
    addressof,
    c_char,
    c_long,
    create_string_buffer,
)
from datetime import datetime
from logging import Logger
from threading import Lock
from xml.etree import cElementTree as eTree

from cfgtools import Group
from modelcommon.electronicpayment import TransactionProcessors

from messagebus import DataType, DefaultToken, Message
from mbcontextmessagehandler import MbContextMessageBus
from modelcommon.customexception import (
    SitefUnavailableException,
)
from modelcommon.model import SitefResponse
from modelcommon.sitef import (
    PaymentCodes,
    SitefFunctions,
    ResultCodes,
    CommandCodes,
    FieldCodes,
    FinalizeActionFunction,
    CancelCodes,
)
from pos_model import TenderType

from paymentprocessor import PaymentProcessor
from typing import (
    Any,
    List,
    Optional,
    Union,
    Dict,
)
from ._SitefActions import SitefActions
from ._SitefParams import SitefParams
from ._SitefServiceFinder import SitefServiceFinder
from ._SitefTPag import SitefTPag
from dialogmanager import DialogManager
from model import (
    CancelPaymentOrderData,
    ConfirmFinishPaymentData,
    FinishPaymentData,
    PaymentResponse,
    PaymentResponseData,
    PaymentResponseMessage,
    StartPaymentData,
    TestConnectionData,
    TPag,
)
from util import (
    get_transaction_identifier, Cache,
)

TK_STORECFG_GET = 0x01300001
TK_SLCTRL_OMGR_ORDERPICT = 0x00A00002

API_SUCCESS_OK = 0
API_LOOP_OK = 10000
API_LOOP_FINISHED = 0
API_NO_SITEF_COMMUNICATION = -5
MINIMUM_TIMEOUT = 15
OPERATOR_CANCEL = -2


def _get_terminal_id(
    pdv_tef     # type: int
):
    # type: (...) -> str

    return "SW%06d" % pdv_tef


def _api_has_data(command):
    return command.value == API_SUCCESS_OK


def _format_api_error_response(api_ret):
    error_responses = {
        -2: PaymentResponseMessage("Cancelado"),
        -5: PaymentResponseMessage("Falha no Servidor SiTEF"),
        -6: PaymentResponseMessage("Operacao cancelada pelo usuario"),
        -40: PaymentResponseMessage("Nao Autorizado - SiTEF"),
        -43: PaymentResponseMessage("Erro PinPad"),
        -120: PaymentResponseMessage("Cancelado")
    }

    if api_ret > 0:
        return PaymentResponseMessage("Nao Autorizado")

    return error_responses.get(api_ret, PaymentResponseMessage("Falha na Operacao: Erro {}".format(api_ret)))


def _should_ignore_command(command):
    return command.value in (2, 4, 21, 14, 52)


def _get_dll_path(config):
    dll_path = config.find_value("DLLPath", "")
    for var in re.findall("{[a-zA-Z0-9]*}", dll_path):
        dll_path = dll_path.replace(var, os.environ.get(var.replace('{', '').replace('}', ''), ""))

    return os.path.normpath(os.path.join(dll_path, "clisitef32i.dll"))


class SitefPaymentProcessor(PaymentProcessor):
    def __init__(
        self,
        message_bus,        # type: MbContextMessageBus
        dialog_manager,     # type: DialogManager
        config,             # type: Group
        logger,             # type: Logger
        cache,              # type: Cache
    ):
        # type: (...) -> None

        super(SitefPaymentProcessor, self).__init__(message_bus, dialog_manager, config, logger, cache)

        from ctypes import WinDLL
        self.logger = logger
        self.dialog_manager = dialog_manager
        self.message_bus = message_bus

        self.store_id = config.find_value("IdLoja")
        self.timeout = int(config.find_value("Timeout", 60))
        if self.timeout < MINIMUM_TIMEOUT:
            self.timeout = MINIMUM_TIMEOUT
        self.connection_test_interval = int(config.find_value("ConnectionTestInterval", 0))
        self.show_qr_code = config.find_value("ShowQRCode", "False").upper() == "TRUE"
        self.is_sync = config.find_value("IsMobileSynch", "False") == "True"
        self.pdv_tef = int(config.find_value("PdvTef") or config.find_value("PosId"))
        self.pos_id = int(config.find_value("PosId"))

        dll_path = _get_dll_path(config)
        self.sitef_dll = WinDLL(dll_path)
        self.configura_sitef_interativo = self.sitef_dll["ConfiguraIntSiTefInterativo"]
        self.inicia_sitef_interativo = self.sitef_dll["IniciaFuncaoSiTefInterativo"]
        self.continua_sitef_interativo = self.sitef_dll["ContinuaFuncaoSiTefInterativo"]
        self.finaliza_sitef_interativo = self.sitef_dll["FinalizaFuncaoSiTefInterativo"]

        self.service_finder = SitefServiceFinder(
            connection_test_interval=self.connection_test_interval,
            sitef_ip_list=config.find_values("IP") or [],
            store_id=self.store_id,
            pdv_tef=self.pdv_tef,
            logger=self.logger,
            dll_path=dll_path,
        )

        self.current_ip = None
        self.current_order = None

        self.cancel = False
        self.cancel_lock = Lock()
        self.get_documents_lock = Lock()

        self.close_dialog_commands = (
            SitefActions.CLEAN_OPERATOR_MESSAGE,
            SitefActions.CLEAN_CUSTOMER_MESSAGE,
            SitefActions.CLEAN_OPERATOR_AND_CUSTOMER_MESSAGE,
        )

        self.show_message_commands = (
            SitefActions.SHOW_OPERATOR_MESSAGE,
            SitefActions.SHOW_OPERATOR_AND_CUSTOMER_MESSAGE,
        )

    def stop(self):  # type: () -> None
        self.service_finder.stop_search_thread()

    def connection_status(
        self,
        test_connection_data    # type: TestConnectionData
    ):
        # type: (...) -> bool

        return self.service_finder.process(
            pos_id=test_connection_data.pos_id,
            data_fiscal=test_connection_data.fiscal_date,
            hora_fiscal=test_connection_data.fiscal_time
        )

    def add_payment(self, start_payment_data, additional_params=""):
        self.service_finder.lock_sitef()
        try:
            result = self._process_loop(start_payment_data, additional_params)
        finally:
            self.dialog_manager.close_dialogs(start_payment_data.pos_id)
            self.service_finder.release_sitef()

        return result

    def cancel_payment_order(
        self,                       # type: SitefPaymentProcessor
        cancel_payment_order_data,  # type: CancelPaymentOrderData
        additional_params="",       # type: Optional[str]
    ):
        # type: (...) -> Any[Union[SitefResponse, str]]

        try:
            self.service_finder.lock_sitef()
            return self.execute_cancel_payment_order(
                cancel_payment_order_data=cancel_payment_order_data,
                additional_params=additional_params,
            )
        finally:
            pos_id = cancel_payment_order_data.pos_id
            self.dialog_manager.close_dialogs(pos_id)
            self.service_finder.release_sitef()

    def execute_cancel_payment_order(
        self,
        cancel_payment_order_data,  # type: CancelPaymentOrderData
        additional_params="",       # type: Optional[str]
    ):
        # type: (...) -> Any[Union[SitefResponse, str]]

        pos_id = cancel_payment_order_data.pos_id
        tender_type = cancel_payment_order_data.tender_type
        tender_amount = cancel_payment_order_data.tender_amount
        order_id = cancel_payment_order_data.order_id
        fiscal_date = str(cancel_payment_order_data.fiscal_date)
        fiscal_time = str(cancel_payment_order_data.fiscal_time)
        tender_id = cancel_payment_order_data.tender_id

        self.logger.debug("SitefProcessor Cancel payment order - STARTED")

        self._check_pin_pad(pos_id)
        self._check_sitef_is_available()
        self._notify_start_process(
            pos_id=int(pos_id),
            tender_type=str(tender_type)
        )

        self.cancel = False
        type_code = SitefFunctions.get_cancel_code(tender_type=tender_type)
        amount = "{:,.2f}".format(float(tender_amount)).replace(",", "").replace(".", ",")
        transaction_identifier = get_transaction_identifier(
            order_id=order_id,
            tender_id=tender_id
        )

        result = self.inicia_sitef_interativo(
            type_code,
            amount,
            transaction_identifier,
            fiscal_date,
            fiscal_time,
            "",
            additional_params
        )

        if result == API_LOOP_OK:
            response = SitefResponse()
            response.transaction_identifier = transaction_identifier
            command = c_long(0)
            field_type = c_long(0)
            min_size = c_long(0)
            max_size = c_long(0)
            buffer_size = 32768
            output_buffer = create_string_buffer('\000' * buffer_size)

            last_time = -1
            last_message = ""
            menu_title = ""

            self.current_order = order_id
            while result == API_LOOP_OK:
                with self.cancel_lock:
                    if self.cancel:
                        keep_going = -1 if tender_type == 122 else 2
                    else:
                        keep_going = 0

                result = self.continua_sitef_interativo(
                    addressof(command),
                    addressof(field_type),
                    addressof(min_size),
                    addressof(max_size),
                    output_buffer,
                    buffer_size,
                    keep_going
                )

                if result != API_LOOP_OK:
                    break

                if not _api_has_data(command):
                    self.logger.debug("COMMAND TYPE: {0} - BUFFER: {1}".format(command.value, output_buffer.value))

                    if command.value in self.show_message_commands:
                        key_cancel = "$INSERT_CARD_CANCEL_PAYMENT|{}".format(cancel_payment_order_data.card_number)
                        data = (command, output_buffer, key_cancel, None)
                        last_message = self._call_command_function(*data)
                        if "insira" not in str(output_buffer.value).lower():
                            self.dialog_manager.show_message(pos_id, output_buffer.value)
                    elif command.value == CommandCodes.set_amount_transaction:
                        output_buffer.value = amount
                    elif self._idle_field(command, field_type, FieldCodes.cancel_transaction_date):
                        output_buffer.value = datetime.strptime(fiscal_date, '%Y%m%d').strftime('%d%m%Y')
                    elif self._idle_field(command, field_type, FieldCodes.cancel_document_number):
                        output_buffer.value = str(cancel_payment_order_data.nsu_host)
                    elif self._idle_field(command, field_type, FieldCodes.cancel_card_number):
                        card_bin = self.dialog_manager.show_input(pos_id, output_buffer.value, input_type="text")
                        if card_bin == "CANCELA" or card_bin == "None":
                            output_buffer.value = "0" * max_size.value
                            self.cancel_payment()
                        else:
                            output_buffer.value = str(card_bin)
                    elif command.value == CommandCodes.idle_command:
                        data = (pos_id, order_id, last_time, output_buffer, max_size, last_message)
                        last_time = self._wait_action(*data)
                    elif command.value == SitefActions.MENU_TITLE:
                        menu_title = output_buffer.value
                    elif command.value == SitefActions.CLEAN_MENU_TITLE:
                        menu_title = ""
                    elif command.value == SitefActions.MENU_CONTENT:
                        self._handle_menu_content(
                            pos_id=pos_id,
                            output_buffer=output_buffer,
                            max_size=max_size,
                            menu_title=menu_title,
                            media=cancel_payment_order_data.media,
                            type_code=type_code,
                        )

                if field_type.value != FieldCodes.empty_information:
                    self.logger.debug("FIELD TYPE: {0} - BUFFER: {1}".format(field_type.value, output_buffer.value))
                    response = self._set_response_cancel_data(
                        response=response,
                        field_type=field_type.value,
                        output_buffer=output_buffer,
                    )

            if result == ResultCodes.success:
                self._notify_end_process(pos_id=pos_id)
                confirm = FinalizeActionFunction.confirm
                self.finaliza_sitef_interativo(
                    confirm,
                    transaction_identifier,
                    fiscal_date,
                    fiscal_time,
                    None
                )
                self._set_cache_receipt_data(
                    order_id=cancel_payment_order_data.order_id,
                    tender_id=cancel_payment_order_data.tender_id,
                    receipt_customer=response.customer_receipt,
                    receipt_merchant=response.merchant_receipt,
                )
                self.logger.debug("SitefProcessor Cancel payment order - SUCCESS")
                return response

            cancel = FinalizeActionFunction.cancel
            self.finaliza_sitef_interativo(
                cancel,
                transaction_identifier,
                fiscal_date,
                fiscal_time,
                None
            )
            self._notify_end_process(pos_id=pos_id)

            return "$ERROR_CANCEL_ELECTRONIC_PAYMENTS_ORDER"

    def cancel_payment(
        self,   # type: SitefPaymentProcessor
    ):
        # type: (...) -> str

        with self.cancel_lock:
            self.logger.info("[SITEF] Cancel payment called")
            self.cancel = True

        return ""

    def finish_payment(
        self,
        finish_payment_data     # type: FinishPaymentData
    ):
        # type: (...) -> bool

        self.logger.debug("SiTEF finish payment started")
        order_id = finish_payment_data.order_id
        pos_id = finish_payment_data.pos_id
        tender_id = finish_payment_data.tender_id
        status = int(finish_payment_data.status)
        transaction_identifier = get_transaction_identifier(
            order_id=order_id,
            tender_id=tender_id
        )
        try:
            self.service_finder.lock_sitef()
            if not self.current_order:
                self.logger.debug("No order to finish, ignoring...")
                return True

            if self.current_order != order_id:
                self.logger.error("order to finish = {}, current order = {}".format(order_id, self.current_order))
                return True

            self.logger.debug("will finish with data: {}".format(finish_payment_data))

            self.configura_sitef_interativo(
                str(self.current_ip),
                str(self.store_id),
                _get_terminal_id(self.pdv_tef),
                0
            )
            fiscal_date = str(finish_payment_data.fiscal_date)
            fiscal_time = str(finish_payment_data.fiscal_time)
            self.finaliza_sitef_interativo(status, transaction_identifier, fiscal_date, fiscal_time, None)

            self.current_ip = self.current_order = None
            self.logger.debug("SiTEF finish payment ended")
            return True
        except Exception as _:
            message = "SITEF Finalize - Pos: {0} / Order: {1} / Status: {2} / Coupon: {3}".format(
                pos_id,
                order_id,
                status,
                transaction_identifier
            )
            self.logger.exception(message)
            return False
        finally:
            self.service_finder.release_sitef()

    def confirm_finish_payment(
        self,
        confirm_finish_payment_data     # type: ConfirmFinishPaymentData
    ):
        # type: (...) -> None

        pass

    def _check_finalized_last_order(self, pos_id):
        for _ in xrange(0, 3, 1):
            if self.current_ip and self.current_order:
                order_id = self.current_order
                order = self._get_order_picture(pos_id, order_id)
                if not order:
                    continue
                order = eTree.XML(order).find("Order")
                created_at = order.get("createdAt").replace("-", "").replace(":", "")
                data_fiscal = created_at[:8]
                hora_fiscal = created_at[9:15]
                if order.find("StateHistory/State[@state='PAID']") is not None:
                    status = "1"
                elif order.find("StateHistory/State[@state='VOIDED']") is not None:
                    status = "0"
                else:
                    self.logger.error("Order %s sem status PAID ou VOIDED" % order_id)
                    return
                self.finish_payment(FinishPaymentData(pos_id, order_id, data_fiscal, hora_fiscal, status, ""))
            else:
                return

        self.logger.warning("Impossivel finalizar order %s" % self.current_order)
        self.current_order = self.current_ip = None

    def _get_order_picture(self, pos_id, order_id):
        try:
            pos_id_to_search = pos_id
            if order_id not in ('', None, 0):
                pos_id_to_search = ''
            params = (pos_id_to_search, order_id, 'false')
            reqbuf = '\x00'.join((str(p) for p in params))
            msg = self.message_bus.send_message("ORDERMGR%s" % pos_id, Message(TK_SLCTRL_OMGR_ORDERPICT, DataType.param, reqbuf))
            params = msg.data.split('\x00')
            if msg.token == DefaultToken.TK_SYS_NAK:
                return None
            return params[2]
        except Exception: # noqa
            self.logger.exception("Exception trying to get last order picture")
            return None

    def get_name(
        self,
    ):
        # type: (...) -> str

        self.register_eft_processor(
            pos_id=self.pos_id,
            processor_name=TransactionProcessors.SITEF,
        )

        return TransactionProcessors.SITEF

    def get_document(
        self,
        pos_id,     # type: str
    ):
        # type: (...) -> str

        self.logger.debug("SiTEF get doc started")
        try:
            self.service_finder.lock_sitef()
            ip_sitef = self.service_finder.find_sitef_service()

            ret = self.configura_sitef_interativo(
                str(self.current_ip),
                str(self.store_id),
                _get_terminal_id(self.pdv_tef),
                0
            )

            if ret != 0:
                self.service_finder.sitef_unavailable(ip_sitef)
                raise Exception("Servidor SiTEF Inoperante")

            fiscal_date = str(datetime.now().date())
            fiscal_time = str(datetime.now().time())

            api_ret = self.inicia_sitef_interativo(789, "", '{}'.format(pos_id), fiscal_date, fiscal_time, "", "")

            if api_ret != 10000:
                raise Exception("Erro ao iniciar SiTEF")

            voided = False
            last_time = -1
            command = c_long(0)
            field_type = c_long(0)
            tam_min = c_long(0)
            tam_max = c_long(0)
            tam_buffer = 32768
            output_buffer = create_string_buffer('\000' * tam_buffer)

            while api_ret == API_LOOP_OK:
                api_ret = self.continua_sitef_interativo(
                    addressof(command), addressof(field_type),
                    addressof(tam_min), addressof(tam_max), output_buffer,
                    tam_buffer, 0
                )

                if api_ret != API_LOOP_OK:
                    break

                if not command.value == 0:
                    if command.value == SitefActions.REQUEST_USER_INFO_WITH_LIMITS:
                        if field_type.value == SitefParams.MESSAGE_CODE:
                            output_buffer.value = "7"

                        elif field_type.value == SitefParams.MIN_SIZE:
                            output_buffer.value = "11"

                        elif field_type.value == SitefParams.MAX_SIZE:
                            output_buffer.value = "11"

                        elif field_type.value == SitefParams.TIME_OUT:
                            output_buffer.value = "0"

                        if last_time == -1:
                            last_time = time.time()
                        else:
                            current_time = time.time()
                            if current_time - last_time > self.timeout:
                                with self.get_documents_lock:
                                    voided = True

                                self.logger.debug(
                                    "SintefGetDoc - Cancel Requested - IP SiTef: {}".format(ip_sitef)
                                )
                                output_buffer.value = "0" * tam_max.value
                    elif command.value == SitefActions.SHOW_CUSTOM_MESSAGE:
                        self.dialog_manager.show_message(pos_id, "$REQUEST_DOCUMENT_ON_PINPAD")
                    elif command.value in self.close_dialog_commands:
                        self.dialog_manager.close_dialogs(pos_id)
                    elif command.value == SitefActions.CANCEL_OPERATION:
                        if field_type.value != SitefParams.CANCEL_OPERATION_REASON:
                            voided = True
                        output_buffer.value = "0"
                        time.sleep(1.5)
                    else:
                        self.dialog_manager.show_message(
                            pos_id, "Mensagem {} {}".format(command.value, output_buffer.value)
                        )
                        last_time = -1


            if api_ret == 0:
                self.finaliza_sitef_interativo("0", '{}'.format(pos_id), fiscal_date, fiscal_time, None)
                self.logger.debug("SiTEF get doc ended")
                return output_buffer.value
            else:
                if voided:
                    raise Exception("Cancelado")

                if api_ret == -5:
                    self.service_finder.sitef_unavailable(ip_sitef)
                    raise Exception("Erro de comunicacao no PinPad: {}".format(api_ret))
                else:
                    raise Exception("Erro na chamada do PinPad: {}".format(api_ret))

        except Exception as ex:
            self.logger.exception("Erro no acesso ao PinPad: {}".format(ex))

        finally:
            self.service_finder.release_sitef()

    def get_last_receipt(
        self,
        eft_processor,      # type: str
    ):
        # type: (...) -> Optional[str]

        return self.cache.get_last_receipt(eft_processor=TransactionProcessors.SITEF)

    def _check_pin_pad(
        self,
        pos_id,                 # type: int
        electronic_type=-1,     # type: int
    ):
        # type: (...) -> None

        if electronic_type == TenderType.digital_wallet:
            return

        self.dialog_manager.show_message(pos_id, "Conectando PinPad")
        time_limit = (time.time() + 5.0)
        while (time.time() < time_limit) and (not self.sitef_dll["VerificaPresencaPinPad"]()):
            time.sleep(0.2)

    def _process_loop(
        self,                    # type: SitefPaymentProcessor
        sitef_loop_data,         # type: StartPaymentData
        additional_params=""     # type: str
    ):
        # type: (...) -> PaymentResponse

        self.logger.debug("SiTEF loop started")
        self.cancel = False
        pos_id = sitef_loop_data.pos_id
        order_id = sitef_loop_data.order_id
        ip_tef = sitef_loop_data.ip_tef
        tender_id = sitef_loop_data.tender_id

        self._check_pin_pad(pos_id, sitef_loop_data.electronic_type)

        if self.current_order == order_id:
            ip_sitef = self.current_ip
        else:
            ip_sitef = self.service_finder.find_sitef_service() if ip_tef is None else ip_tef
            self._check_finalized_last_order(pos_id)

        if not ip_sitef:
            self.service_finder.wakeup_search()
            self.logger.warning("No IP available")
            raise SitefUnavailableException("Sem conexao SITEF")

        self._check_sitef_is_available()

        amount = sitef_loop_data.amount
        electronic_type = sitef_loop_data.electronic_type

        payment_code = SitefFunctions.get_payment_code(electronic_type=electronic_type)
        is_digital_wallets = payment_code == PaymentCodes.digital_wallet

        transaction_identifier = get_transaction_identifier(
            order_id=order_id,
            tender_id=tender_id,
        )
        fiscal_date = str(sitef_loop_data.fiscal_date)
        fiscal_time = str(sitef_loop_data.fiscal_time)
        operator = str(sitef_loop_data.operator)
        api_ret = self.inicia_sitef_interativo(
            payment_code,
            amount,
            transaction_identifier,
            fiscal_date,
            fiscal_time,
            operator,
            additional_params
        )

        if api_ret != API_LOOP_OK:
            self._log_api_start_ret_error(api_ret)
            return PaymentResponseMessage("Pagamento interrompido")

        self.current_order = order_id
        self.current_ip = ip_sitef

        command = c_long(0)
        field_type = c_long(0)
        min_size = c_long(0)
        max_size = c_long(0)
        buffer_size = 32768
        output_buffer = create_string_buffer('\000' * buffer_size)

        voided = False

        institution_name = ""
        card_bin = None
        transaction_date = None
        auth_code = None
        authorizer_name = None
        authorizer_cnpj = None
        card_type = None
        expiration_date = None
        nsu = None
        nsu_host = None
        receipt_merchant = None
        receipt_customer = None
        id_mercado_pago = None
        menu_title = ""
        last_digits = "0000"
        sat_acquirer = ""
        t_band = ""

        show_message_commands = (
            SitefActions.SHOW_OPERATOR_MESSAGE,
            SitefActions.SHOW_OPERATOR_AND_CUSTOMER_MESSAGE,
        )

        while api_ret == API_LOOP_OK:
            keep_going = 0
            with self.cancel_lock:
                if self.cancel:
                    keep_going = -1 if is_digital_wallets else 2

            api_ret = self.continua_sitef_interativo(
                addressof(command),
                addressof(field_type),
                addressof(min_size),
                addressof(max_size),
                output_buffer,
                buffer_size,
                keep_going
            )

            log_message = "API RET {0} - COMMAND TYPE: {1} - FIELD TYPE {2} - BUFFER: {3}"
            self.logger.debug(log_message.format(str(api_ret), command.value, field_type.value, output_buffer.value))
            if api_ret != API_LOOP_OK:
                break

            if not _api_has_data(command) and not self._must_ignore_command(field_type):
                if command.value in show_message_commands:
                    self.dialog_manager.show_message(pos_id, output_buffer.value)
                elif command.value == SitefActions.SHOW_MESSAGE_WITH_CONFIRMATION:
                    self.dialog_manager.show_question(pos_id, output_buffer.value, buttons="$OK")
                elif command.value == SitefActions.SHOW_MESSAGE_WITH_CANCEL_BUTTON:
                    self.dialog_manager.show_message(pos_id, output_buffer.value)
                elif command.value == SitefActions.ASK_PIN_CODE:
                    input_type = "text" if sitef_loop_data.electronic_type == TenderType.gift_card else "number"
                    pin = self.dialog_manager.show_input(pos_id, output_buffer.value, input_type=input_type)
                    if pin == "CANCELA" or pin == "None":
                        self.cancel_payment()
                        output_buffer.value = "0" * max_size.value
                    else:
                        output_buffer.value = str(pin)
                elif command.value == SitefActions.CANCEL_OPERATION:
                    output_buffer.value = "0"
                    time.sleep(1.5)

                    if field_type.value != 111:
                        voided = True
                elif command.value == SitefActions.MENU_TITLE:
                    menu_title = output_buffer.value
                elif command.value == SitefActions.CLEAN_MENU_TITLE:
                    menu_title = ""
                elif command.value == SitefActions.MENU_CONTENT:
                    choices = {option.split(':')[0]: option.split(':')[1] for option in output_buffer.value.split(';') if option}
                    selected_choice = self.dialog_manager.show_choices(pos_id, choices, message=menu_title)
                    if str(selected_choice) == "-1":
                        self.cancel_payment()
                    else:
                        output_buffer.value = str(selected_choice)

                elif command.value == SitefActions.CLOSE_QRCODE:
                    self.dialog_manager.close_qrcode(pos_id)
                elif _should_ignore_command(command):
                    time.sleep(0.5)
                else:
                    self.dialog_manager.show_message(pos_id, "Mensagem {} {}".format(command.value, output_buffer.value))

            self.logger.debug("FIELD TYPE: {0} - VALUE: {1}".format(field_type.value, output_buffer.value))
            if field_type.value == SitefParams.CARD_BIN:
                card_bin = output_buffer.value
            elif field_type.value == SitefParams.CARD_TYPE:
                card_type = output_buffer.value
            elif field_type.value in SitefParams.AUTH_CODE:
                auth_code = output_buffer.value
            elif self._is_show_qrcode_command(field_type):
                self.dialog_manager.show_qrcode(pos_id, output_buffer.value)
            elif field_type.value == SitefParams.RECEIPT_CUSTOMER:
                receipt_customer = output_buffer.value
            elif field_type.value == SitefParams.RECEIPT_MERCHANT:
                receipt_merchant = output_buffer.value
            elif field_type.value == SitefParams.NSU:
                nsu = output_buffer.value
            elif field_type.value == SitefParams.NSU_HOST:
                nsu_host = output_buffer.value
            elif field_type.value == SitefParams.TRANSACTION_DATE:
                transaction_date = output_buffer.value
            elif field_type.value == SitefParams.AUTHORIZER_NAME:
                authorizer_name = output_buffer.value
            elif field_type.value == SitefParams.INSTITUTION_NAME:
                institution_name = output_buffer.value
            elif field_type.value == SitefParams.EXPIRATION_DATE:
                expiration_date = output_buffer.value
            elif field_type.value == SitefParams.AUTHORIZER_CNPJ:
                authorizer_cnpj = output_buffer.value
            elif field_type.value == SitefParams.ID_MERCADO_PAGO:
                id_mercado_pago = output_buffer.value
            elif field_type.value == SitefParams.DIGITAL_WALLET_ID:
                card_type = output_buffer.value
            elif field_type.value == SitefParams.LAST_CARD_DIGITS:
                last_digits = output_buffer.value
            elif field_type.value == SitefParams.SAT_ACQUIRER:
                sat_acquirer = output_buffer.value
            elif field_type.value == SitefParams.NFCE_BRAND:
                t_band = output_buffer.value

        if api_ret == API_LOOP_FINISHED and auth_code is not None:
            self.logger.debug("SiTEF loop finished")

            receipt_customer = self._encode_receipt(
                receipt=receipt_customer,
            )
            receipt_merchant = self._encode_receipt(
                receipt=receipt_merchant,
            )

            self._set_cache_receipt_data(
                order_id=order_id,
                tender_id=tender_id,
                receipt_customer=receipt_customer,
                receipt_merchant=receipt_merchant,
            )

            t_pag = SitefTPag().get_t_pag(
                card_type=card_type,
                payment_code=payment_code,
                electronic_type=electronic_type,
            )

            return PaymentResponseData(
                media=institution_name.strip(),
                result=api_ret,
                card_number=card_bin,
                auth_code="{}{}".format(transaction_date, id_mercado_pago) if is_digital_wallets else auth_code,
                approved_amount=amount,
                id_auth=authorizer_name,
                cnpj_auth=authorizer_cnpj,
                bandeira=card_type,
                receipt_merchant=receipt_merchant,
                receipt_customer=receipt_customer,
                expiration_date=expiration_date,
                nsu=nsu,
                transaction_processor=TransactionProcessors.SITEF,
                last_digits=last_digits,
                sat_acquirer=sat_acquirer,
                t_pag=t_pag,
                t_band=t_band,
                nsu_host=nsu_host,
                electronic_type=tender_id,
                transaction_identifier=transaction_identifier
            )

        if voided:
            return PaymentResponseMessage("Cancelado")

        if api_ret == OPERATOR_CANCEL:
            self.logger.info("[SITEF] Operator cancel transaction")
            self._cancel_pending_transaction(
                pos_id=pos_id,
                order_id=order_id,
                transaction_identifier=transaction_identifier,
                fiscal_date=fiscal_date,
                fiscal_time=fiscal_time,
            )
            self.logger.info("[SITEF] End process cancel transaction")

        if api_ret == API_NO_SITEF_COMMUNICATION:
            self.service_finder.sitef_unavailable(ip_sitef)

        return _format_api_error_response(api_ret)

    def _encode_receipt(
        self,
        receipt,    # type: str
    ):
        # type: (...) -> str

        try:
            receipt = base64.b64encode(receipt)
        except Exception as _:
            self.logger.exception("Error encoding receipt: {}".format(receipt))
            receipt = ''

        return receipt

    def _must_ignore_command(self, field_type):
        return self._is_show_qrcode_command(field_type)

    @staticmethod
    def _is_show_qrcode_command(field_type):
        return field_type.value == SitefParams.QR_CODE

    def _log_api_config_ret_error(self, api_ret):
        error_codes = {
            0: "No Error",
            1: "IP Error",
            2: "Store Code Error",
            3: "Terminal Code Error",
            6: "TCP/IP Error",
            7: "Memory Error",
            8: "CliSiTef Error",
            10: "CliSiTef Access Error"
        }
        self.logger.error("API Config Error: {}".format(error_codes.get(api_ret, "Unknown")))

    def _log_api_start_ret_error(self, api_ret):
        self.logger.warning("API Start Error: {}".format(api_ret))

    @staticmethod
    def _get_coupon_code(
        pos_id,     # type: int
        order_id,   # type: int
        tender_id   # type: str
    ):
        return "{0}{1}{2}".format(pos_id, order_id, tender_id)

    def _check_sitef_is_available(
        self
    ):
        # type: (...) -> None

        self.current_ip = self.service_finder.find_sitef_service() if self.current_ip is None else self.current_ip
        result = self.configura_sitef_interativo(
            str(self.current_ip),
            str(self.store_id),
            _get_terminal_id(self.pdv_tef)
            ,
            0
        )
        if result != API_SUCCESS_OK:
            self.service_finder.sitef_unavailable(self.current_ip)
            self._log_api_config_ret_error(result)
            raise SitefUnavailableException("$SITEF_UNAVAILABLE")

    def _notify_start_process(
        self,
        pos_id,         # type: int
        tender_type     # type: str
    ):
        # type: (...) -> None

        self.message_bus.mbcontext.MB_EasyEvtSend(
            subject="SITEF",
            type="PROCESS_STARTED",
            xml=tender_type,
            sourceid=pos_id,
            synchronous=True
        )

    def _notify_status_update(
        self,
        pos_id,     # type: Any[Union[str, int]]
        data=""     # type: Optional[str]
    ):
        # type: (...) -> None

        data = ";".join([data, pos_id])
        self.message_bus.mbcontext.MB_EasyEvtSend(
            subject="SITEF_STATUS_UPDATE",
            type="STATUS_UPDATE",
            xml=data
        )

    def _notify_end_process(
        self,
        pos_id  # type: int
    ):
        # type: (...) -> None

        self.message_bus.mbcontext.MB_EasyEvtSend(
            subject="SITEF_STATUS_UPDATE_FINISHED",
            type="SITEF_STATUS_UPDATE_FINISHED",
            xml=str(pos_id)
        )

    @staticmethod
    def _idle_field(
        command,        # type: c_long
        field_type,     # type: c_long,
        value           # type: int
    ):
        # type: (...) -> bool

        return command.value == CommandCodes.idle_field and field_type.value == value

    def _wait_action(
        self,
        pos_id,         # type: str
        order_id,       # type: str
        last_time,      # type: int
        output_buffer,  # type: Array[c_char]
        max_size,       # type: c_long
        last_message    # type: str
    ):
        # type: (...) -> float

        if last_time == -1:
            return time.time()

        current_time = time.time()
        if current_time - last_time > self.timeout:
            with self.cancel_lock:
                self.cancel = True

            log_message = "SyncSitefProcessor - Wait Action - OrderId: {0} - IP SiTef: {1}"
            self.logger.debug(log_message.format(order_id, self.current_ip))

            output_buffer.value = "0" * max_size.value
            return last_time

        delimiter = "|" if last_message.startswith("$") else " "
        message = "{}{}{}".format(last_message, delimiter, str(self.timeout - int(current_time - last_time)))
        self._notify_status_update(
            pos_id=pos_id,
            data=message
        )

        return last_time

    def _call_command_function(
        self,
        command,                    # type: c_long
        output_buffer,              # type: Array[c_char]
        key_translate=None,         # type: str
        extra_data_display=None     # type: Optional[List[str]]
    ):
        # type: (...) -> str

        if command.value in [CommandCodes.show_message_operator, CommandCodes.show_message_both]:
            return self.format_message_translate(
                output_buffer=output_buffer,
                key_translate=key_translate,
                extra_data_display=extra_data_display
            )

        if command.value in [CommandCodes.clear_message_operator, CommandCodes.clear_message_both]:
            # TODO: In SITEF here was a call to SitefCallback
            return ""

        return ""

    def _handle_menu_content(
        self,           # type: SitefPaymentProcessor
        pos_id,         # type: str
        output_buffer,  # type: Array[c_char]
        max_size,       # type: c_long
        menu_title,     # type: str
        media,          # type: str
        type_code,      # type: int
    ):
        # type: (...) -> None

        choices = self._convert_output_buffer_to_choices(options=output_buffer.value)

        if type_code == CancelCodes.digital_wallet:
            for key, value in choices.items():
                if value == media:
                    output_buffer.value = key
                    return

        selected_choice = self._show_choice_options(
            pos_id=pos_id,
            menu_title=menu_title,
            choices=choices,
        )

        if selected_choice is None:
            output_buffer.value = "0" * max_size.value
            self.cancel_payment()
            return

        output_buffer.value = selected_choice

    @staticmethod
    def _convert_output_buffer_to_choices(
        options,    # type: str
    ):
        # type: (...) -> Dict

        options = options.split(';')

        return {option.split(':')[0]: option.split(':')[1] for option in options if option}

    def _show_choice_options(
        self,           # type: SitefPaymentProcessor
        pos_id,         # type: str
        menu_title,     # type: str
        choices,        # type: Dict
    ):
        # type: (...) -> Optional[str]

        selected_choice = self.dialog_manager.show_choices(
            pos_id=pos_id,
            choices=choices,
            message=menu_title,
        )

        if selected_choice == "-1":
            return None

        return selected_choice

    @staticmethod
    def format_message_translate(
        output_buffer,              # type: Array[c_char]
        key_translate,              # type: Optional[str]
        extra_data_display=None     # type: Optional[List[str]]
    ):
        # type: (...) -> str

        if key_translate is not None and extra_data_display is None:
            return key_translate

        if extra_data_display is None:
            return output_buffer.value

        extra_data_display.insert(0, output_buffer.value)
        message = "|".join([str(e) for e in extra_data_display])

        return "{}|{}".format(key_translate, message)

    @staticmethod
    def _set_response_cancel_data(
        response,       # type: SitefResponse
        field_type,     # type: int
        output_buffer,  # type: Array[c_char]
    ):
        # type: (...) -> SitefResponse

        if field_type == FieldCodes.receipt_customer:
            response.customer_receipt = base64.b64encode(output_buffer.value)
        elif field_type == FieldCodes.receipt_merchant:
            response.merchant_receipt = base64.b64encode(output_buffer.value)
        elif field_type == FieldCodes.nsu:
            response.nsu = output_buffer.value
        elif field_type == FieldCodes.nsu_host:
            response.nsu_host = output_buffer.value
        elif field_type == FieldCodes.nsu_origin:
            response.nsu_origin = output_buffer.value
        elif field_type == FieldCodes.transaction_value:
            response.transaction_value = output_buffer.value

        return response

    def _set_cache_receipt_data(
        self,
        order_id,           # type: str
        tender_id,          # type: str
        receipt_customer,   # type: str
        receipt_merchant,   # type: str
    ):
        # type: (...) -> None

        receipt_data = {
            "order_id": order_id,
            "tender_id": tender_id,
            "receipt_customer": receipt_customer,
            "receipt_merchant": receipt_merchant,
        }
        receipt_data = json.dumps(receipt_data)
        self.cache.set_last_receipt(
            receipt_data=receipt_data,
            eft_processor=TransactionProcessors.SITEF,
        )

    def _cancel_pending_transaction(
        self,                       # type: SitefPaymentProcessor
        pos_id,                     # type: str
        order_id,                   # type: str
        transaction_identifier,     # type: str
        fiscal_date,                # type: str
        fiscal_time,                # type: str
    ):
        # type: (...) -> None

        message = "[SITEF] Cancel pending transaction. Pos id: {} - Order Id: {} - Transaction: {}"
        self.logger.info(message, pos_id, order_id, transaction_identifier)

        finish_payment_data = FinishPaymentData(
            pos_id=pos_id,
            order_id=order_id,
            fiscal_date=fiscal_date,
            fiscal_time=fiscal_time,
            status="0",
            tender_details="{}",
        )
        self.finish_payment(finish_payment_data=finish_payment_data)

        message = "[SITEF] Finish cancel pending transaction. Pos id: {} - Order Id: {} - Transaction: {}"
        self.logger.info(message, pos_id, order_id, transaction_identifier)

    def get_qr_code_config(
        self,   # type: SitefPaymentProcessor
    ):
        # type: (...) -> bool

        return self.show_qr_code




from datetime import datetime


class Order(object):
    def __init__(
        self,
        order_id,   # type: int
        date,       # type: datetime
        state,      # type: int
    ):
        # type: (...) -> None

        self.order_id = order_id
        self.date = date
        self.state = state



import datetime
from io import BytesIO

from botocore.auth import (
    _host_from_url, _get_body_as_dict, BaseSigner,
    SIGNED_HEADERS_BLACKLIST, UNSIGNED_PAYLOAD
)
from botocore.compat import awscrt, HTTPHeaders, parse_qs, urlsplit, urlunsplit
from botocore.utils import percent_encode_sequence
from botocore.exceptions import NoCredentialsError

class CrtSigV4Auth(BaseSigner):
    REQUIRES_REGION = True
    _PRESIGNED_HEADERS_BLACKLIST = [
        'Authorization',
        'X-Amz-Date',
        'X-Amz-Content-SHA256',
        'X-Amz-Security-Token',
    ]
    _SIGNATURE_TYPE = awscrt.auth.AwsSignatureType.HTTP_REQUEST_HEADERS
    _USE_DOUBLE_URI_ENCODE = True
    _SHOULD_NORMALIZE_URI_PATH = True

    def __init__(self, credentials, service_name, region_name):
        self.credentials = credentials
        self._service_name = service_name
        self._region_name = region_name
        self._expiration_in_seconds = None

    def add_auth(self, request):
        if self.credentials is None:
            raise NoCredentialsError()

        # Use utcnow() because that's what gets mocked by tests, but set
        # timezone because CRT assumes naive datetime is local time.
        datetime_now = datetime.datetime.utcnow().replace(
            tzinfo=datetime.timezone.utc)

        # Use existing 'X-Amz-Content-SHA256' header if able
        existing_sha256 = self._get_existing_sha256(request)

        self._modify_request_before_signing(request)

        credentials_provider = awscrt.auth.AwsCredentialsProvider.new_static(
            access_key_id=self.credentials.access_key,
            secret_access_key=self.credentials.secret_key,
            session_token=self.credentials.token)

        if self._should_sha256_sign_payload(request):
            if existing_sha256:
                explicit_payload = existing_sha256
            else:
                explicit_payload = None  # to be calculated during signing
        else:
            explicit_payload = UNSIGNED_PAYLOAD

        if self._should_add_content_sha256_header(explicit_payload):
            body_header = \
                awscrt.auth.AwsSignedBodyHeaderType.X_AMZ_CONTENT_SHA_256
        else:
            body_header = awscrt.auth.AwsSignedBodyHeaderType.NONE

        signing_config = awscrt.auth.AwsSigningConfig(
            algorithm=awscrt.auth.AwsSigningAlgorithm.V4,
            signature_type=self._SIGNATURE_TYPE,
            credentials_provider=credentials_provider,
            region=self._region_name,
            service=self._service_name,
            date=datetime_now,
            should_sign_header=self._should_sign_header,
            use_double_uri_encode=self._USE_DOUBLE_URI_ENCODE,
            should_normalize_uri_path=self._SHOULD_NORMALIZE_URI_PATH,
            signed_body_value=explicit_payload,
            signed_body_header_type=body_header,
            expiration_in_seconds=self._expiration_in_seconds,
            )
        crt_request = self._crt_request_from_aws_request(request)
        future = awscrt.auth.aws_sign_request(crt_request, signing_config)
        future.result()
        self._apply_signing_changes(request, crt_request)

    def _crt_request_from_aws_request(self, aws_request):
        url_parts = urlsplit(aws_request.url)
        crt_path = url_parts.path if url_parts.path else '/'
        if aws_request.params:
            array = []
            for (param, value) in aws_request.params.items():
                value = str(value)
                array.append('%s=%s' % (param, value))
            crt_path = crt_path + '?' + '&'.join(array)
        elif url_parts.query:
            crt_path = '%s?%s' % (crt_path, url_parts.query)

        crt_headers = awscrt.http.HttpHeaders(aws_request.headers.items())

        # CRT requires body (if it exists) to be an I/O stream.
        crt_body_stream = None
        if aws_request.body:
            if hasattr(aws_request.body, 'seek'):
                crt_body_stream = aws_request.body
            else:
                crt_body_stream = BytesIO(aws_request.body)

        crt_request = awscrt.http.HttpRequest(
            method=aws_request.method,
            path=crt_path,
            headers=crt_headers,
            body_stream=crt_body_stream)
        return crt_request

    def _apply_signing_changes(self, aws_request, signed_crt_request):
        # Apply changes from signed CRT request to the AWSRequest
        aws_request.headers = HTTPHeaders.from_pairs(
            list(signed_crt_request.headers))

    def _should_sign_header(self, name, **kwargs):
        return name.lower() not in SIGNED_HEADERS_BLACKLIST

    def _modify_request_before_signing(self, request):
        # This could be a retry. Make sure the previous
        # authorization headers are removed first.
        for h in self._PRESIGNED_HEADERS_BLACKLIST:
            if h in request.headers:
                del request.headers[h]
        # If necessary, add the host header
        if 'host' not in request.headers:
            request.headers['host'] = _host_from_url(request.url)

    def _get_existing_sha256(self, request):
        return request.headers.get('X-Amz-Content-SHA256')

    def _should_sha256_sign_payload(self, request):
        # Payloads will always be signed over insecure connections.
        if not request.url.startswith('https'):
            return True

        # Certain operations may have payload signing disabled by default.
        # Since we don't have access to the operation model, we pass in this
        # bit of metadata through the request context.
        return request.context.get('payload_signing_enabled', True)

    def _should_add_content_sha256_header(self, explicit_payload):
        # only add X-Amz-Content-SHA256 header if payload is explicitly set
        return explicit_payload is not None


class CrtS3SigV4Auth(CrtSigV4Auth):
    # For S3, we do not normalize the path.
    _USE_DOUBLE_URI_ENCODE = False
    _SHOULD_NORMALIZE_URI_PATH = False

    def _get_existing_sha256(self, request):
        # always recalculate
        return None

    def _should_sha256_sign_payload(self, request):
        # S3 allows optional body signing, so to minimize the performance
        # impact, we opt to not SHA256 sign the body on streaming uploads,
        # provided that we're on https.
        client_config = request.context.get('client_config')
        s3_config = getattr(client_config, 's3', None)

        # The config could be None if it isn't set, or if the customer sets it
        # to None.
        if s3_config is None:
            s3_config = {}

        # The explicit configuration takes precedence over any implicit
        # configuration.
        sign_payload = s3_config.get('payload_signing_enabled', None)
        if sign_payload is not None:
            return sign_payload

        # We require that both content-md5 be present and https be enabled
        # to implicitly disable body signing. The combination of TLS and
        # content-md5 is sufficiently secure and durable for us to be
        # confident in the request without body signing.
        if not request.url.startswith('https') or \
                'Content-MD5' not in request.headers:
            return True

        # If the input is streaming we disable body signing by default.
        if request.context.get('has_streaming_input', False):
            return False

        # If the S3-specific checks had no results, delegate to the generic
        # checks.
        return super()._should_sha256_sign_payload(request)

    def _should_add_content_sha256_header(self, explicit_payload):
        # Always add X-Amz-Content-SHA256 header
        return True


class CrtSigV4QueryAuth(CrtSigV4Auth):
    DEFAULT_EXPIRES = 3600
    _SIGNATURE_TYPE = awscrt.auth.AwsSignatureType.HTTP_REQUEST_QUERY_PARAMS

    def __init__(self, credentials, service_name, region_name,
                 expires=DEFAULT_EXPIRES):
        super().__init__(credentials, service_name, region_name)
        self._expiration_in_seconds = expires

    def _modify_request_before_signing(self, request):
        super()._modify_request_before_signing(request)

        # We automatically set this header, so if it's the auto-set value we
        # want to get rid of it since it doesn't make sense for presigned urls.
        content_type = request.headers.get('content-type')
        if content_type == 'application/x-www-form-urlencoded; charset=utf-8':
            del request.headers['content-type']

        # Now parse the original query string to a dict, inject our new query
        # params, and serialize back to a query string.
        url_parts = urlsplit(request.url)
        # parse_qs makes each value a list, but in our case we know we won't
        # have repeated keys so we know we have single element lists which we
        # can convert back to scalar values.
        query_dict = dict(
            [(k, v[0]) for k, v in
             parse_qs(url_parts.query, keep_blank_values=True).items()])
        # The spec is particular about this.  It *has* to be:
        # https://<endpoint>?<operation params>&<auth params>
        # You can't mix the two types of params together, i.e just keep doing
        # new_query_params.update(op_params)
        # new_query_params.update(auth_params)
        # percent_encode_sequence(new_query_params)
        if request.data:
            # We also need to move the body params into the query string. To
            # do this, we first have to convert it to a dict.
            query_dict.update(_get_body_as_dict(request))
            request.data = ''
        new_query_string = percent_encode_sequence(query_dict)
        # url_parts is a tuple (and therefore immutable) so we need to create
        # a new url_parts with the new query string.
        # <part>   - <index>
        # scheme   - 0
        # netloc   - 1
        # path     - 2
        # query    - 3  <-- we're replacing this.
        # fragment - 4
        p = url_parts
        new_url_parts = (p[0], p[1], p[2], new_query_string, p[4])
        request.url = urlunsplit(new_url_parts)

    def _apply_signing_changes(self, aws_request, signed_crt_request):
        # Apply changes from signed CRT request to the AWSRequest
        super()._apply_signing_changes(aws_request, signed_crt_request)

        signed_query = urlsplit(signed_crt_request.path).query
        p = urlsplit(aws_request.url)
        # urlsplit() returns a tuple (and therefore immutable) so we
        # need to create new url with the new query string.
        # <part>   - <index>
        # scheme   - 0
        # netloc   - 1
        # path     - 2
        # query    - 3  <-- we're replacing this.
        # fragment - 4
        aws_request.url = urlunsplit((p[0], p[1], p[2], signed_query, p[4]))


class CrtS3SigV4QueryAuth(CrtSigV4QueryAuth):
    """S3 SigV4 auth using query parameters.
    This signer will sign a request using query parameters and signature
    version 4, i.e a "presigned url" signer.
    Based off of:
    http://docs.aws.amazon.com/AmazonS3/latest/API/sigv4-query-string-auth.html
    """

    # For S3, we do not normalize the path.
    _USE_DOUBLE_URI_ENCODE = False
    _SHOULD_NORMALIZE_URI_PATH = False

    def _should_sha256_sign_payload(self, request):
        # From the doc link above:
        # "You don't include a payload hash in the Canonical Request, because
        # when you create a presigned URL, you don't know anything about the
        # payload. Instead, you use a constant string "UNSIGNED-PAYLOAD".
        return False

    def _should_add_content_sha256_header(self, explicit_payload):
        # Never add X-Amz-Content-SHA256 header
        return False


# Defined at the bottom of module to ensure all Auth
# classes are defined.
CRT_AUTH_TYPE_MAPS = {
    'v4': CrtSigV4Auth,
    'v4-query': CrtSigV4QueryAuth,
    's3v4': CrtS3SigV4Auth,
    's3v4-query': CrtS3SigV4QueryAuth,
}



# -*- coding: utf-8 -*-


class RuptureEvents(object):
    FullRuptureRequest = "FullRuptureRequest"
    CleanRupture = "CleanRupture"
    OrderModified = "ORDER_MODIFIED"


class RuptureEventTypes(object):
    CleanRuptureThread = "CleanRuptureThread"
    CleanRuptureOnOpenDay = "OnOpenDay"



# edb.py - functions for handling North Macedonia EDB numbers
# coding: utf-8
#
# Copyright (C) 2022 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""ЕДБ (Едниствен Даночен Број, North Macedonia tax number).

This number consists of 13 digits, sometimes with an additional "MK" prefix.

More information:

* http://www.ujp.gov.mk/en

>>> validate('4030000375897')
'4030000375897'
>>> validate('МК 4020990116747')  # Cyrillic letters
'4020990116747'
>>> validate('MK4057009501106')  # ASCII letters
'4057009501106'
>>> validate('4030000375890')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('МК 4020990116747')  # Cyrillic letters
'4020990116747'
>>> format('MK4057009501106')  # ASCII letters
'4057009501106'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace.
    """
    number = clean(number, ' -').upper().strip()
    # First two are ASCII, second two are Cyrillic and only strip matching
    # types to avoid implicit conversion to unicode strings in Python 2.7
    for prefix in ('MK', u'MK', 'МК', u'МК'):
        if isinstance(number, type(prefix)) and number.startswith(prefix):
            number = number[len(prefix):]
    return number


def calc_check_digit(number):
    """Calculate the check digit."""
    weights = (7, 6, 5, 4, 3, 2, 7, 6, 5, 4, 3, 2)
    total = sum(int(n) * w for n, w in zip(number, weights))
    return str((-total % 11) % 10)


def validate(number):
    """Check if the number is a valid North Macedonia ЕДБ number.

    This checks the length, formatting and check digit.
    """
    number = compact(number)
    if len(number) != 13:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if number[-1] != calc_check_digit(number):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid North Macedonia ЕДБ number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number)



import json
import logging
import time

from typing import (
    List,
    Optional,
)

from bustoken import (
    TK_NDISCOUNT_ADD_BENEFIT,
    TK_NDISCOUNT_GET_BENEFITS,
    TK_NDISCOUNT_REMOVE,
)
from msgbus import (
    FM_PARAM,
    MBMessage,
    TK_SYS_ACK,
)
from sysactions import send_message

from model.customexception import (
    DiscountMinLevelException,
    ErrorApplyingDiscount,
    ErrorGettingBenefits,
)


class NDiscountInteractor(object):
    def __init__(self):
        # type: () -> None
        self.logger = logging.getLogger("PosActions")

    def get_benefits(
        self,           # type: NDiscountInteractor
        pos_id,         # type: int
        only_visible,   # type: bool
    ):
        # type: (...) -> json

        self.logger.info("Getting benefits for POS: {}".format(pos_id))
        retry = 0
        while retry < 3:
            time.sleep(0.5 * retry)
            try:
                data = json.dumps(dict(posId=pos_id, only_visible=only_visible))
                response = send_message(
                    dest_name="nDiscount",
                    token=TK_NDISCOUNT_GET_BENEFITS,
                    format=FM_PARAM,
                    data=data,
                    timeout=30000000,
                )
                if response.token == TK_SYS_ACK:
                    return json.loads(response.data)
            except Exception as ex:
                self.logger.warn("Error getting benefits data: {} - Retry attempt: {}".format(ex.message, retry))

            retry += 1

        self.logger.exception("[get_benefits] benefits retry failed")
        raise ErrorGettingBenefits()

    @staticmethod
    def add_benefit(
            pos_id,                     # type: str
            benefit_id,                 # type: str
            benefit,                    # type: json
            user_auth_level=False       # type: Optional[int]
    ):
        # type: (...) -> None

        data = json.dumps(dict(
                posId=pos_id,
                benefitId=benefit_id,
                benefit=benefit,
                user_auth_level=user_auth_level
        ))

        response = send_message("nDiscount", TK_NDISCOUNT_ADD_BENEFIT, FM_PARAM, data, timeout=30000000)
        if response.token != TK_SYS_ACK:
            error_code = response.data.split("\0")[0]
            if error_code == "NDISCOUNT_ERROR_USER_MIN_LEVEL":
                raise DiscountMinLevelException(min_level_required=int(response.data.split("\0")[1]))

            raise ErrorApplyingDiscount(error_code=response.data)

    @staticmethod
    def remove_benefits(pos_id, benefit_ids=None):
        # type: (str, List) -> MBMessage

        data = json.dumps(dict({"pos_id": pos_id, "benefit": benefit_ids}))
        return send_message("nDiscount", TK_NDISCOUNT_REMOVE, FM_PARAM, data, timeout=30000000)

    @staticmethod
    def remove_benefit(pos_id, benefit_id):
        # type: (str, int) -> MBMessage

        data = json.dumps(dict({"pos_id": pos_id, "benefit": benefit_id}))
        return send_message("nDiscount", TK_NDISCOUNT_REMOVE, FM_PARAM, data, timeout=30000000)

    @staticmethod
    def test_connection():
        # type: () -> bool
        try:
            data = json.dumps(dict())
            response = send_message("nDiscount", TK_NDISCOUNT_GET_BENEFITS, FM_PARAM, data, timeout=30000000)
            if response:
                return True
            return False
        except Exception as _:
            return False



# -*- coding: utf-8 -*-

from abc import abstractmethod, ABCMeta
from old_helper import round_half_away_from_zero

from comp_exceptions import FiscalParametersException


class TaxBuilder(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def calculate_icms(self, cst_icms, csosn, tax_item, base_reduction, product_code):
        pass

    @abstractmethod
    def calculate_pis(self, cst_pis, tax_item):
        pass

    @abstractmethod
    def calculate_cofins(self, cst_cofins, tax_item):
        pass

    @staticmethod
    def calculate_approximated_tax_rate(ibtp_tax_processor, sale_item):
        item_approximated_tax = ibtp_tax_processor.get_item_tax(sale_item)
        return "<vItem12741>{}</vItem12741>".format(item_approximated_tax)


class TaxBuilderLucroReal(TaxBuilder):
    def __init__(self):
        super(TaxBuilder, self).__init__()

    ICMS_00 = "<ICMS>" \
                "<ICMS00>" \
                  "<Orig>{orig}</Orig>" \
                  "<CST>{cst:>02}</CST>" \
                  "<pICMS>{p_icms:.02f}</pICMS>" \
                "</ICMS00>" \
              "</ICMS>"

    ICMS_40 = "<ICMS>" \
                "<ICMS40>" \
                  "<Orig>{orig}</Orig>" \
                  "<CST>{cst:>02}</CST>" \
                "</ICMS40>" \
              "</ICMS>"

    PIS_NT = "<PIS>" \
               "<PISNT>" \
                 "<CST>{cst:>02}</CST>" \
               "</PISNT>" \
             "</PIS>"

    PIS_ALIQ = "<PIS>" \
                 "<PISAliq>" \
                   "<CST>{cst:>02}</CST>" \
                   "<vBC>{v_bc:.02f}</vBC>" \
                   "<pPIS>{p_pis:.04f}</pPIS>" \
                 "</PISAliq>" \
               "</PIS>"

    COFINS_NT = "<COFINS>" \
                  "<COFINSNT>" \
                    "<CST>{cst:>02}</CST>" \
                  "</COFINSNT>" \
                "</COFINS>"

    COFINS_ALIQ = "<COFINS>" \
                    "<COFINSAliq>" \
                      "<CST>{cst:>02}</CST>" \
                      "<vBC>{v_bc:.02f}</vBC>" \
                      "<pCOFINS>{p_cofins:.04f}</pCOFINS>" \
                    "</COFINSAliq>" \
                  "</COFINS>"

    def calculate_icms(self, cst_icms, csosn, tax_item, base_reduction, product_code):
        _round_tax_amount(tax_item)
        if cst_icms in (0, 90) and tax_item:
            xml = self.ICMS_00.format(orig=0, cst=cst_icms, p_icms=tax_item.tax_rate * 100)
            return xml, tax_item.base_amount_ad, tax_item.tax_amount_ad
        elif cst_icms == 20 and tax_item:
            tax_base_reduction = (1 - base_reduction / 100) * tax_item.tax_rate * 100
            xml = self.ICMS_00.format(orig=0, cst=cst_icms, p_icms=tax_base_reduction)
            return xml, tax_item.base_amount_ad, tax_item.tax_amount_ad
        elif cst_icms in (40, 41, 50, 60):
            return self.ICMS_40.format(orig=0, cst=cst_icms), 0, 0
        else:
            if not tax_item:
                raise FiscalParametersException("Sem ICMS para produto {}".format(product_code))
            raise FiscalParametersException("CST ICMS/Taxa Desconhecido {} Produto {}".format(cst_icms, product_code))

    def calculate_pis(self, cst_pis, tax_item):
        _round_tax_amount(tax_item)
        if cst_pis in (1, 2, 99) and tax_item:
            xml = self.PIS_ALIQ.format(cst=cst_pis, v_bc=tax_item.base_amount_ad, p_pis=tax_item.tax_rate)
            return xml, tax_item.base_amount_ad, tax_item.tax_amount_ad

        elif cst_pis in (4, 5, 6, 7, 8, 9):
            return self.PIS_NT.format(cst=cst_pis), 0, 0

        else:
            raise FiscalParametersException("CST PIS/Taxa Desconhecido %s" % cst_pis)

    def calculate_cofins(self, cst_cofins, tax_item):
        _round_tax_amount(tax_item)
        if cst_cofins in (1, 2, 99) and tax_item:
            xml = self.COFINS_ALIQ.format(cst=cst_cofins, v_bc=tax_item.base_amount_ad, p_cofins=tax_item.tax_rate)
            return xml, tax_item.base_amount_ad, tax_item.tax_amount_ad

        elif cst_cofins in (4, 5, 6, 7, 8, 9):
            return self.COFINS_NT.format(cst=cst_cofins), 0, 0

        else:
            raise FiscalParametersException("CST COFINS Desconhecido %s" % cst_cofins)


class TaxBuilderSimplesNacional(TaxBuilder):
    def __init__(self):
        super(TaxBuilderSimplesNacional, self).__init__()

    ICMSSN_500 = "<ICMS>" \
                   "<ICMSSN500>" \
                     "<Orig>{orig}</Orig>" \
                     "<CSOSN>{csosn}</CSOSN>" \
                   "</ICMSSN500>" \
                 "</ICMS>"

    ICMSSN_102 = "<ICMS>" \
                   "<ICMSSN102>" \
                     "<Orig>{orig}</Orig>" \
                     "<CSOSN>{csosn}</CSOSN>" \
                   "</ICMSSN102>" \
                 "</ICMS>"

    PIS_SN = "<PIS>" \
               "<PISSN>" \
                 "<CST>{cst:0>2}</CST>" \
               "</PISSN>"\
              "</PIS>"

    PIS_NT = "<PIS>" \
               "<PISNT>" \
                 "<CST>{cst:>02}</CST>" \
               "</PISNT>" \
             "</PIS>"

    COFINS_SN = "<COFINS>" \
                  "<COFINSSN>" \
                    "<CST>{cst:0>2}</CST>" \
                  "</COFINSSN>"\
                "</COFINS>"

    COFINS_NT = "<COFINS>" \
                  "<COFINSNT>" \
                    "<CST>{cst:>02}</CST>" \
                  "</COFINSNT>" \
                "</COFINS>"

    def calculate_icms(self, cst_icms, csosn, tax_item, base_reduction, product_code):
        if csosn in [102, 300, 400, 500]:
            return self.ICMSSN_102.format(orig=0, csosn=csosn), 0, 0
        else:
            raise FiscalParametersException("CSOSN Desconhecido {} Produto {}".format(csosn, product_code))

    def calculate_pis(self, cst_pis, tax_item):
        if cst_pis in (1, 2, 49, 99):
            return self.PIS_SN.format(cst=49), 0, 0
        elif cst_pis in (4, 5, 6, 7, 8, 9):
            return self.PIS_NT.format(cst=cst_pis), 0, 0
        else:
            raise FiscalParametersException("CST PIS/Taxa Desconhecido %s" % cst_pis)

    def calculate_cofins(self, cst_cofins, tax_item):
        if cst_cofins in (1, 2, 49, 99):
            return self.COFINS_SN.format(cst=49), 0, 0
        elif cst_cofins in (4, 5, 6, 7, 8, 9):
            return self.COFINS_NT.format(cst=cst_cofins), 0, 0
        else:
            raise FiscalParametersException("CST COFINS Desconhecido %s" % cst_cofins)


class TaxBuilderSimplesNacionalSublimiteReceita(TaxBuilder):
    def __init__(self):
        super(TaxBuilderSimplesNacionalSublimiteReceita, self).__init__()

    ICMS_00 = "<ICMS>" \
                "<ICMS00>" \
                  "<Orig>{orig}</Orig>" \
                  "<CST>{cst:>02}</CST>" \
                  "<pICMS>{p_icms:.02f}</pICMS>" \
                  "</ICMS00>" \
                "</ICMS>"

    ICMS_40 = "<ICMS>" \
                "<ICMS40>" \
                  "<Orig>{orig}</Orig>" \
                  "<CST>{cst:>02}</CST>" \
                "</ICMS40>" \
              "</ICMS>"

    PIS_OUTR = "<PIS>" \
                 "<PISOutr>" \
                   "<CST>{cst:0>2}</CST>" \
                   "<qBCProd>{q_bc_prod:.02f}</qBCProd>" \
                   "<vAliqProd>{v_aliq_prod:.04f}</vAliqProd>" \
                 "</PISOutr>" \
               "</PIS>"

    PIS_NT = "<PIS>" \
               "<PISNT>" \
                 "<CST>{cst:>02}</CST>" \
               "</PISNT>" \
             "</PIS>"

    COFINS_OUTR = "<COFINS>" \
                    "<COFINSOutr>" \
                      "<CST>{cst:0>2}</CST>" \
                      "<qBCProd>{q_bc_prod:.02f}</qBCProd>" \
                      "<vAliqProd>{v_aliq_prod:.04f}</vAliqProd>" \
                    "</COFINSOutr>" \
                  "</COFINS>"

    COFINS_NT = "<COFINS>" \
                  "<COFINSNT>" \
                    "<CST>{cst:>02}</CST>" \
                  "</COFINSNT>" \
                "</COFINS>"

    def calculate_icms(self, cst_icms, csosn, tax_item, base_reduction, product_code):
        _round_tax_amount(tax_item)

        if cst_icms in (0, 20, 90):
            tax_rate = tax_item.tax_rate * 100
            return self.ICMS_00.format(orig=0, cst=0, p_icms=tax_rate), tax_item.base_amount_ad, tax_item.tax_amount_ad
        elif cst_icms in (40, 41, 50, 60):
            return self.ICMS_40.format(orig=0, cst=cst_icms), 0, 0
        else:
            if not tax_item:
                raise FiscalParametersException("Sem ICMS para produto {}".format(product_code))
            raise FiscalParametersException("CST ICMS/Taxa Desconhecido {} Produto {}".format(cst_icms, product_code))

    def calculate_pis(self, cst_pis, tax_item):
        if cst_pis in (1, 2, 99):
            return self.PIS_OUTR.format(cst=99, q_bc_prod=0, v_aliq_prod=0), 0, 0
        elif cst_pis in (4, 5, 6, 7, 8, 9):
            return self.PIS_NT.format(cst=cst_pis), 0, 0
        else:
            raise FiscalParametersException("CST PIS/Taxa Desconhecido %s" % cst_pis)

    def calculate_cofins(self, cst_cofins, tax_item):
        if cst_cofins in (1, 2, 99):
            return self.COFINS_OUTR.format(cst=99, q_bc_prod=0, v_aliq_prod=0), 0, 0
        elif cst_cofins in (4, 5, 6, 7, 8, 9):
            return self.COFINS_NT.format(cst=cst_cofins), 0, 0
        else:
            raise FiscalParametersException("CST COFINS Desconhecido %s" % cst_cofins)


def _round_tax_amount(tax_item):
    if tax_item:
        tax_item.base_amount_ad = round_half_away_from_zero(tax_item.base_amount_ad, 2)
        tax_item.tax_amount_ad = round_half_away_from_zero(tax_item.tax_amount_ad, 2)



from application.interactor import (
    GetLoyaltyCustomerInfoInteractor,
    LockBenefitInteractor,
    LoyaltyAvailableInteractor,
    LoyaltyIdentifierInteractor,
    GetAndLockCouponInteractor,
)
from application.interactor.burn_voucher import BurnVoucherInteractor
from application.interactor.unlock_voucher import UnlockVoucherInteractor
from application.model.configuration import Configurations
from application.model.customexception import (
    ClientCustomError,
    NotFoundError,
    InUseError,
    GoneError,
)
from application.processor import (
    GetLoyaltyCustomerInfoProcessor,
    BurnVoucherProcessor,
    UnlockVoucherProcessor,
    LoyaltyAvailableProcessor,
    LoyaltyIdentifierProcessor,
)
from application.processor.lock_benefit import LockBenefitProcessor
from application.processor.tokens import ListenedToken
from application.processor.get_and_lock_coupon import GetAndLockCouponProcessor
from application.repository.api import APIRepository
from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Message, Event
from messagehandler import MessageHandlerBuilder
from messageprocessor import (
    MessageProcessorMessageHandler,
    DefaultMessageProcessorExecutorFactory,
)
from messageprocessorutil import LoggingProcessorCallback
from typing import Union


class ComponentMessageHandlerBuilder(MessageHandlerBuilder):

    def __init__(
        self,
        configs,        # type: Configurations
        message_bus,    # type: MbContextMessageBus
    ):
        # type: (...) -> None

        self.logger = configs.logger
        self.configs = configs
        self.api_repository = APIRepository(
            configs=self.configs,
            message_bus=message_bus,
        )

        self.message_handler = None  # type: Union[None, MessageProcessorMessageHandler]

    def build_singletons(self):
        # type: () -> None

        get_loyalty_customer_info_interactor = GetLoyaltyCustomerInfoInteractor(
            api_repository=self.api_repository,
        )
        get_and_lock_coupon_interactor = GetAndLockCouponInteractor(api_repository=self.api_repository)
        lock_benefit_interactor = LockBenefitInteractor(api_repository=self.api_repository)
        burn_voucher_interactor = BurnVoucherInteractor(api_repository=self.api_repository)
        unlock_voucher_interactor = UnlockVoucherInteractor(api_repository=self.api_repository)
        loyalty_available_interactor = LoyaltyAvailableInteractor(self.configs)
        loyalty_identifier_interactor = LoyaltyIdentifierInteractor(configs=self.configs)

        get_loyalty_customer_info_processor = GetLoyaltyCustomerInfoProcessor(
            interactor=get_loyalty_customer_info_interactor,
        )
        get_and_lock_voucher_processor = GetAndLockCouponProcessor(interactor=get_and_lock_coupon_interactor)
        lock_benefit_processor = LockBenefitProcessor(interactor=lock_benefit_interactor)
        burn_voucher_processor = BurnVoucherProcessor(interactor=burn_voucher_interactor)
        unlock_voucher_processor = UnlockVoucherProcessor(interactor=unlock_voucher_interactor)
        loyalty_available_processor = LoyaltyAvailableProcessor(loyalty_available_interactor)
        loyalty_identifier_processor = LoyaltyIdentifierProcessor(interactor=loyalty_identifier_interactor)

        messages_processors = {
            ListenedToken.TK_LOYALTY_GET_LOYALTY_ID.value: get_loyalty_customer_info_processor,
            ListenedToken.TK_LOYALTY_GET_AND_LOCK_VOUCHER.value: get_and_lock_voucher_processor,
            ListenedToken.TK_LOYALTY_LOCK_BENEFIT.value: lock_benefit_processor,
            ListenedToken.TK_LOYALTY_BURN_VOUCHER.value: burn_voucher_processor,
            ListenedToken.TK_LOYALTY_UNLOCK_VOUCHER.value: unlock_voucher_processor,
            ListenedToken.TK_LOYALTY_IS_AVAILABLE.value: loyalty_available_processor,
            ListenedToken.TK_LOYALTY_GET_ID_TYPE.value: loyalty_identifier_processor,
        }

        callbacks = [
            LoggingProcessorCallback(
                logger=self.logger,
                ignored_exceptions=[
                    ClientCustomError,
                    NotFoundError,
                    InUseError,
                    GoneError,
                ]
            )
        ]

        self.message_handler = MessageProcessorMessageHandler(
            event_processors=None,
            message_processors=messages_processors,
            executor_factory=DefaultMessageProcessorExecutorFactory(callbacks),
        )

    def build_message_handler(self, message, event):
        # type: (Message, Event) -> None

        return self.message_handler

    def destroy_singletons(self):
        # type: () -> None

        return

    def destroy_message_handler(self, message_handler):
        # type: (MessageHandlerBuilder) -> None

        return



# -*- coding: utf-8 -*-

import logging
import os
import re
import xml.etree.ElementTree as eTree

import cfgtools
import pyscripts
import sysactions
from cfgtools import Configuration
from helper import import_pydevd, remove_accents
from messagehandler import MessageHandler
from modelcommon.model import Store
from msgbus import MBEasyContext
from old_helper import OrderTaker, PosUtil, config_logger, read_swconfig, remove_punctuation
from pos_model import OrderParser
from pos_util import SaleLineUtil
from systools import sys_parsefilepath
from typing import (
    Union,
    Dict,
    List,
    Any,
)

import nfcebuilder
import satbuilder
from aposprinter import SatAposA8FiscalPrinter, NfceAposA8FiscalPrinter
from clichehelper import ClicheHelper
from common import FiscalParameterController, TipBuilderCommon, PrinterCommon, CalculateTaxesItemCommon
from eventhandler import FiscalWrapperEventHandler
from ibpttaxprocessor import IbptTaxProcessor
from models import (
    SoftwareDeveloper,
    LoyaltyConfigurations,
)
from nfce import NfceProcessor, NfceRequestBuilder, NfceContingencia, NfceAutorizador, NfceConnectivityTester, \
    NfceSituationChecker
from nfcebuilder.nfceutil import NfeSigner, XmlEnveloper, NfceRequest
from nfcecanceler import NfceCanceler, NfceCancelerRequestBuilder, NfceCancelerParameters, NfceCancelerResponseParser
from nfcedisabler import OrderDisabler, NfceDisabler, NfceDisablerRequestBuilder, NfceDisablerParameters, \
    NfceDisablerResponseParser, OrderRepository, ProcInutNfeBuilder, ProcInutNfeSaver
from order_retriever import OrderRetrieverService, OrderDataFormatter
from order_retriever.repository import OrderRetrieverRepository
from printer import SatFiscalPrinter, NfceFiscalPrinter, DonationCouponPrinter
from repository import FiscalParameterRepository, SequencerRepository, ProductRepository
from satbuilder import CfeBuilder
from services import SequencerService
from sat import SatProcessor, SatRequestBuilder, SatServiceFinder, MfeServiceFinder
from operator import itemgetter

REQUIRED_SERVICES = "FiscalPersistence|PosController|StoreWideConfig"


def main():
    loader_cfg = os.environ["LOADERCFG"]
    import_pydevd(loader_cfg, 9125)

    config = cfgtools.read(loader_cfg)
    service_name = str(config.find_value("Identification.ComponentName"))
    logger = logging.getLogger(service_name)

    config_logger(loader_cfg, service_name)
    config_logger(loader_cfg, service_name + "Thread")
    config_logger(loader_cfg, service_name + "NfceDisabler")
    config_logger(loader_cfg, service_name + "NfceCanceler")
    config_logger(loader_cfg, service_name + "Coupons")

    xml_enviados = str(config.find_value("FiscalWrapper.XmlEnviados"))
    xml_erros = str(config.find_value("FiscalWrapper.XmlErros"))
    xml_inconsistentes = str(config.find_value("FiscalWrapper.XmlInconsistentes") or "Inconsistentes")
    export_xml_path = os.path.join(sys_parsefilepath(str(config.find_value("FiscalWrapper.XmlPath"))), service_name)
    crt = int(config.find_value("FiscalWrapper.crt"))
    ibpt_tax_files_path = sys_parsefilepath(str(config.find_value("FiscalWrapper.IbptTaxFilesPath") or "../data/bundles/fiscalwrapper/ibpttaxfiles"))

    mbcontext = MBEasyContext(service_name)
    pyscripts.mbcontext = mbcontext
    sysactions.mbcontext = mbcontext

    nf_type = config.find_value("FiscalWrapper.NFType")
    order_ref_ids = _get_order_ref_ids(config)
    order_ref_position = (config.find_value("FiscalWrapper.OrderReferencePosition") or "BETWEEN_FISCAL_DATA").upper()

    fiscal_printer = None
    mobile_fiscal_printer = None
    fiscal_processor = None

    fiscal_sent_dir = sys_parsefilepath(config.find_value("FiscalWrapper.FiscalSentDir").strip())

    message_handler = MessageHandler(mbcontext, service_name, service_name, REQUIRED_SERVICES)

    nome_emit = remove_accents(read_swconfig(mbcontext, "Store.RazaoSocial"))
    end_logradouro = _limit_max_length(remove_accents(read_swconfig(mbcontext, "Store.EndLogradouro")), 60)
    end_numero = remove_punctuation(read_swconfig(mbcontext, "Store.EndNumero").strip()) or "S/N"
    end_compl = remove_accents((read_swconfig(mbcontext, "Store.EndCompl") or "").strip())
    bairro = remove_accents(read_swconfig(mbcontext, "Store.Bairro"))
    municipio = remove_accents(read_swconfig(mbcontext, "Store.Municipio"))
    cep = remove_punctuation(read_swconfig(mbcontext, "Store.CEP"))
    uf = remove_accents(read_swconfig(mbcontext, "Store.UF"))
    cnpj_contribuinte = remove_punctuation(config.find_value("FiscalWrapper.CnpjContribuinte"))
    inscr_estadual = remove_punctuation(config.find_value("FiscalWrapper.InscEstadual"))

    store_model = Store(
        crt,
        nome_emit,
        cnpj_contribuinte,
        inscr_estadual,
        cep,
        end_logradouro,
        end_numero,
        end_compl,
        bairro,
        municipio,
        uf
    )

    cliche_helper = ClicheHelper(mbcontext, cnpj_contribuinte, inscr_estadual)

    order_disabler = None
    nfce_contingencia = None
    nfce_situation_checker = None
    nfce_connectivity_tester = None
    nfce_canceler = None
    versao_ws = None
    synchronous_mode = False

    loader_fiscal_parameters = _get_config_group(config, "FiscalWrapper.FiscalParameters")
    fiscal_coupon_configurations = _get_fiscal_coupon_configurations(config=config)
    fiscal_parameters = FiscalParameterRepository(mbcontext).get_fiscal_parameters(loader_fiscal_parameters)
    tenders_type = ProductRepository(mbcontext).get_tender_types()

    customer_text_field = config.find_value("FiscalWrapper.PrinterCustomizations.CustomerTextField") or ""
    print_order_identifier = (config.find_value("FiscalWrapper.PrinterCustomizations.PrintOrderIdentifier") or "false").lower() == "true"
    customer_call_by_field = config.find_value("FiscalWrapper.PrinterCustomizations.YouWillBeCalledBy") or ""

    period_to_retry_orders_with_exception = config.find_value("FiscalWrapper.PeriodToRetryOrdersWithException") or "7"

    data_path = sys_parsefilepath("{HVDATADIR}")
    database_dir = os.path.join(data_path, "databases")

    default_fiscal_path = "../../src/persistcomp/fiscal/fiscal.ddl"
    fiscal_ddl_path = config.find_value("FiscalWrapper.FiscalSchemaDatabasePath", default=default_fiscal_path)
    fiscal_schema_sql_path = os.path.join(data_path, fiscal_ddl_path)

    order_parser = OrderParser()

    tip_config = _get_tip_config(config)
    tip_config_range = _get_tip_config_by_range(config, fiscal_parameters)
    calculate_taxes_common = CalculateTaxesItemCommon(mbcontext)
    fiscal_max_value = float(config.find_value("FiscalWrapper.MaxValueToFiscalize") or "10000")
    fiscal_parameter_controller = FiscalParameterController(
        fiscal_parameters=fiscal_parameters,
        tip_config_range=tip_config_range,
        loader_fiscal_parameters=loader_fiscal_parameters,
        fiscal_max_value=fiscal_max_value,
    )
    tip_builder_common = TipBuilderCommon(
                    tip_config,
                    tip_config_range,
                    mbcontext,
                    fiscal_parameter_controller,
                    calculate_taxes_common
    )
    printer_common = PrinterCommon(
        tip_config=tip_config,
        tip_config_range=tip_config_range,
        tip_builder_common=tip_builder_common,
        customer_call_by_field=customer_call_by_field,
    )
    loyalty_configurations = _get_loyalty_configurations(
        mb_context=mbcontext,
        config=config,
    )

    ibtp_tax_processor = IbptTaxProcessor(mbcontext, ibpt_tax_files_path, uf, fiscal_parameter_controller)

    sale_line_util = SaleLineUtil()

    order_retriever_repository = OrderRetrieverRepository(mbcontext)
    order_data_formatter = OrderDataFormatter()
    order_retriever_service = OrderRetrieverService(order_retriever_repository, order_data_formatter)

    initial_order_id = None
    font_size_withdrawal_panel = _get_font_size_withdrawal_panel_mobile(
        config=config,
    )

    danfe_font_size_mobile = _get_danfe_font_size_mobile(
      config=config,
    )

    if nf_type.upper() == "SAT" or nf_type.upper() == "MFE":
        sat_group = config.find_group("FiscalWrapper.SAT")
        cnpj_sw = sat_group.find_value("CnpjSW")
        sign_ac = sat_group.find_value("SignAC")
        max_sat_number = int(sat_group.find_value("MaxSatNumber"))
        pos_sat_list = sat_group.find_value("PosSatList")
        sleep_time = int(sat_group.find_value("SearchSatInterval"))
        layout_version = sat_group.find_value("LayoutVersion", default="0.07")
        is_mfe = nf_type.upper() == "MFE"

        if nf_type.upper() == "SAT":
            sat_service_finder = SatServiceFinder(mbcontext, max_sat_number, pos_sat_list, sleep_time)
        else:
            sat_service_finder = MfeServiceFinder(mbcontext, max_sat_number, pos_sat_list, sleep_time)

        cfe_builder = _build_sat_builders(
            crt=crt,
            cnpj_contribuinte=cnpj_contribuinte,
            inscr_estadual=inscr_estadual,
            cnpj_sw=cnpj_sw,
            sign_ac=sign_ac,
            fiscal_parameter_controller=fiscal_parameter_controller,
            sale_line_util=sale_line_util,
            tip_config=tip_config,
            ibtp_tax_processor=ibtp_tax_processor,
            customer_text_field=customer_text_field,
            tenders_type=tenders_type,
            tip_builder_common=tip_builder_common,
            layout_version=layout_version,
        )

        sat_request_builder = SatRequestBuilder(
            mbcontext=mbcontext,
            crt=crt,
            cnpj_sw=cnpj_sw,
            sign_ac=sign_ac,
            cnpj_contribuinte=cnpj_contribuinte,
            inscr_estadual=inscr_estadual,
            order_parser=order_parser,
            fiscal_parameter_controller=fiscal_parameter_controller,
            sale_line_util=sale_line_util,
            cfe_builder=cfe_builder,
            is_mfe=is_mfe,
        )
        fiscal_processor = SatProcessor(mbcontext, sat_service_finder, sat_request_builder, fiscal_sent_dir)
        fiscal_printer = SatFiscalPrinter(
            mb_context=mbcontext,
            order_ref_ids=order_ref_ids,
            cliche_helper=cliche_helper,
            customer_text_field=customer_text_field,
            print_order_identifier=print_order_identifier,
            printer_common=printer_common,
            coupon_fiscal_parameters=fiscal_coupon_configurations,
            order_ref_position=order_ref_position,
            loyalty_configurations=loyalty_configurations,
        )
        mobile_fiscal_printer = SatAposA8FiscalPrinter(
            mb_context=mbcontext,
            order_ref_ids=order_ref_ids,
            cliche_helper=cliche_helper,
            customer_text_field=customer_text_field,
            printer_common=printer_common,
            coupon_fiscal_parameters=fiscal_coupon_configurations,
            font_size_withdrawal_panel=font_size_withdrawal_panel,
            danfe_font_size_mobile=danfe_font_size_mobile,
            order_ref_position=order_ref_position,
            loyalty_configurations=loyalty_configurations,
        )

    elif nf_type.upper() in ("NFCE", "PAF"):
        nfce_group = config.find_group("FiscalWrapper.NFCE")

        c_uf = nfce_group.find_value("cUF")
        c_mun_fg = nfce_group.find_value("cMunFG")
        mod = nfce_group.find_value("mod")
        serie = nfce_group.find_value("serie")
        ambiente = nfce_group.find_value("ambiente")
        csc = nfce_group.find_value("csc")
        cid_token = nfce_group.find_value("cid_token")
        initial_order_id = int(nfce_group.find_value("initial_order_id"))

        serie_inc, initial_order_id_inc = _get_development_serie_and_order_id_increment()
        if serie_inc is not None:
            serie = str(int(serie) + serie_inc)
            initial_order_id += initial_order_id_inc

        versao_ws = int(nfce_group.find_value("versao_ws"))
        send_sleep_time = int(nfce_group.find_value("send_sleep_time") or 5)
        check_situation_timeout = int(nfce_group.find_value("CheckSituationTimeout") or 10)
        synchronous_mode = (nfce_group.find_value("synchronous_mode") or "false").lower() == "true"

        store_model.city_code = c_mun_fg
        store_model.state_code = c_uf

        software_developer = _read_software_developer_informations(config, uf)

        bundle_dir = os.environ["BUNDLEDIR"]
        certificate_key_path = bundle_dir + nfce_group.find_value("certificate_key_path")
        certificate_path = bundle_dir + nfce_group.find_value("certificate_path")
        nfe_signer = NfeSigner(certificate_key_path, certificate_path)
        sefaz_certificate_path = bundle_dir + nfce_group.find_value("sefaz_certificate_path")

        nfce_urls = _get_nfce_urls(logger, nfce_group, uf, versao_ws, ambiente)

        nfe_builder = _build_nfe_builders(
                serie,
                mod,
                ambiente,
                fiscal_parameter_controller,
                sale_line_util,
                cid_token,
                csc,
                nfce_urls["qrcode_base_url"],
                nfe_signer,
                versao_ws,
                nfce_urls["qrcode_check_url"],
                software_developer,
                ibtp_tax_processor,
                customer_text_field,
                tenders_type,
                tip_builder_common,
                store_model
        )

        nfce_request_builder = NfceRequestBuilder(mbcontext, initial_order_id, crt, cnpj_contribuinte, inscr_estadual,
                                                  c_uf, uf, c_mun_fg, mod, serie, ambiente, end_logradouro, end_numero,
                                                  end_compl, bairro, municipio, cep, nome_emit, csc, cid_token,
                                                  nfce_urls["qrcode_base_url"], certificate_key_path, certificate_path, order_parser,
                                                  fiscal_parameter_controller, sale_line_util, nfe_builder, versao_ws, synchronous_mode)
        nfce_request = NfceRequest(certificate_key_path, certificate_path, sefaz_certificate_path, check_situation_timeout)

        max_tantavivas_envio_lote = int(nfce_group.find_value("max_tantavivas_envio_lote") or "3")
        max_tantavivas_envio_contingencia = int(nfce_group.find_value("max_tantavivas_envio_contingencia") or "10")
        intervalo_retentativa_lote = int(nfce_group.find_value("intervalo_retentativa_lote") or "15")



        if nf_type.upper() == "NFCE":
            nfce_connectivity_tester = NfceConnectivityTester(nfce_request, nfce_urls["url_status_servico"], c_uf, versao_ws)
            nfce_situation_checker = NfceSituationChecker(nfce_request, nfce_urls["url_consulta_nfe"], ambiente, c_uf, versao_ws)
            nfce_autorizador = NfceAutorizador(nfce_request_builder, nfce_request, nfce_urls["url_autorizacao"], nfce_urls["url_ret_autorizacao"], versao_ws, max_tantavivas_envio_lote, intervalo_retentativa_lote, send_sleep_time, synchronous_mode)
            nfce_contingencia = NfceContingencia(
                mbcontext=mbcontext,
                nfce_request_builder=nfce_request_builder,
                nfce_autorizador=nfce_autorizador,
                versao_ws=versao_ws,
                nfce_situation_checker=nfce_situation_checker,
                fiscal_sent_dir=fiscal_sent_dir,
                max_tantavivas_envio_contingencia=max_tantavivas_envio_contingencia,
                synchronous_mode=synchronous_mode,
                nfce_connectivity_tester=nfce_connectivity_tester,
                database_dir=database_dir,
                fiscal_schema_sql_path=fiscal_schema_sql_path,
            )
            fiscal_printer = NfceFiscalPrinter(
                mb_context=mbcontext,
                order_ref_ids=order_ref_ids,
                cliche_helper=cliche_helper,
                qrcode_check_url=nfce_urls["qrcode_check_url"],
                versao_ws=versao_ws,
                customer_text_field=customer_text_field,
                print_order_identifier=print_order_identifier,
                printer_common=printer_common,
                coupon_fiscal_parameters=fiscal_coupon_configurations,
                order_ref_position=order_ref_position,
                loyalty_configurations=loyalty_configurations,
            )
            mobile_fiscal_printer = NfceAposA8FiscalPrinter(
                mb_context=mbcontext,
                order_ref_ids=order_ref_ids,
                cliche_helper=cliche_helper,
                qrcode_url=nfce_urls["qrcode_check_url"],
                customer_text_field=customer_text_field,
                printer_common=printer_common,
                coupon_fiscal_parameters=fiscal_coupon_configurations,
                font_size_withdrawal_panel=font_size_withdrawal_panel,
                danfe_font_size_mobile=danfe_font_size_mobile,
                order_ref_position=order_ref_position,
                loyalty_configurations=loyalty_configurations,
            )
        else:
            nfce_request = None
            nfce_autorizador = None
            nfce_contingencia = None
            fiscal_printer = None
            nfce_situation_checker = None

        fiscal_processor = NfceProcessor(
            mbcontext=mbcontext,
            nfce_request_builder=nfce_request_builder,
            nfce_request=nfce_request,
            nfce_autorizador=nfce_autorizador,
            nfce_contingencia=nfce_contingencia,
            url_sefaz=nfce_urls["qrcode_base_url"],
            fiscal_sent_dir=fiscal_sent_dir,
            versao_ws=versao_ws,
            synchronous_mode=synchronous_mode,
            nfce_situation_checker=nfce_situation_checker,
            xml_enviados=xml_enviados,
        )

        justificativa_inutilizacao = nfce_group.find_value("justificativa_inutilizacao")
        nfce_disabler_parameters = NfceDisablerParameters(initial_order_id, ambiente, c_uf, cnpj_contribuinte, serie, justificativa_inutilizacao)
        nfce_canceler_parameters = NfceCancelerParameters(ambiente, c_uf, cnpj_contribuinte)

        xml_enveloper = XmlEnveloper(c_uf)
        nfce_disabler_request_builder = NfceDisablerRequestBuilder(nfce_disabler_parameters, nfe_signer, xml_enveloper, versao_ws)
        nfce_canceler_request_builder = NfceCancelerRequestBuilder(nfce_canceler_parameters, nfe_signer, xml_enveloper)

        nfce_disabler_response_parser = NfceDisablerResponseParser(versao_ws)
        nfce_canceler_response_parser = NfceCancelerResponseParser()

        nfce_canceler = NfceCanceler(nfce_canceler_request_builder, nfce_canceler_response_parser, nfce_request, nfce_urls["url_cancelamento"], fiscal_sent_dir, versao_ws)

        proc_inut_nfe_builder = ProcInutNfeBuilder()
        proc_inut_nfe_saver = ProcInutNfeSaver(fiscal_sent_dir)
        nfce_disabler = NfceDisabler(nfce_disabler_request_builder, nfce_disabler_response_parser, nfce_request, nfce_urls["url_inutilizacao"], proc_inut_nfe_builder, proc_inut_nfe_saver, versao_ws)

        order_taker = OrderTaker()

        pos_list = PosUtil(mbcontext).get_pos_list()
        order_repository = OrderRepository(mbcontext, pos_list)

        order_disabler = OrderDisabler(order_repository, nfce_disabler, order_taker)

    sequencer_repository = SequencerRepository(mbcontext)
    sequencer_service = SequencerService(sequencer_repository)

    last_order_id = int(order_retriever_service.get_last_order_id())
    initial_order_id = initial_order_id or 0
    sequencer_service.update_fiscal_id_sequence(initial_order_id, last_order_id)

    printers = {
        "": fiscal_printer,
        "APOS_A8": mobile_fiscal_printer,
        "EXTERNAL": mobile_fiscal_printer,
    }
    donation_coupon_printer = DonationCouponPrinter(cliche_helper, nf_type.upper())

    event_handler = FiscalWrapperEventHandler(
        mbcontext=mbcontext,
        fiscal_processor=fiscal_processor,
        printers=printers,
        nfce_connectivity_tester=nfce_connectivity_tester,
        order_disabler=order_disabler,
        nfce_canceler=nfce_canceler,
        fiscal_parameter_controller=fiscal_parameter_controller,
        xml_enviados=xml_enviados,
        xml_erros=xml_erros,
        export_xml_path=export_xml_path,
        fiscal_sent_dir=fiscal_sent_dir,
        nf_type=nf_type,
        versao_ws=versao_ws,
        ibpt_tax_processor=ibtp_tax_processor,
        order_retriever_service=order_retriever_service,
        xml_inconsistentes=xml_inconsistentes,
        donation_coupon_printer=donation_coupon_printer,
        nfce_contingencia=nfce_contingencia,
        period_to_retry_orders_with_exception=period_to_retry_orders_with_exception,
        synchronous_mode=synchronous_mode,
        sequencer_service=sequencer_service,
    )
    message_handler.set_event_handler(event_handler)

    if nfce_contingencia is not None:
        nfce_contingencia.start_contingencia()

    message_handler.subscribe_events([
        "ORDER_MODIFIED",
        "DisableNfceOrder",
        "ReSignXMLs",
        "OrderManagerFiscalXmlRequest",
        "CheckIfFiscalXmlIsCreatedInFolder",
        "RetryFiscalOrdersWithException"
    ])

    message_handler.handle_events()


def _get_nfce_urls(logger, nfce_group, uf, versao_ws, ambiente):
    urls_path = sys_parsefilepath(nfce_group.find_value("urls_path")) or "../data/server/bundles/fiscalwrapper/nfce_urls.xml"
    nfce_urls = {}

    if os.path.exists(urls_path):
        logger.info("Reading NFCe URLs from <nfce_urls.xml>...")
        urls_xml = eTree.parse(urls_path).getroot()
        for state in urls_xml.findall("estado"):
            if state.get("uf") == uf:
                for version in state.findall("versao"):
                    if int(version.get("ws")) == versao_ws and version.get("ambiente") == ambiente:
                        for url in version:
                            nfce_urls[url.tag] = str.strip(url.text)

    if not nfce_urls:
        logger.info("NFCe URLs not found in <nfce_urls.xml>, reading from <loader.cfg>...")
        nfce_urls["url_autorizacao"] = nfce_group.find_value("url_autorizacao") or ""
        nfce_urls["url_ret_autorizacao"] = nfce_group.find_value("url_ret_autorizacao") or ""
        nfce_urls["url_inutilizacao"] = nfce_group.find_value("url_inutilizacao") or ""
        nfce_urls["url_cancelamento"] = nfce_group.find_value("url_cancelamento") or ""
        nfce_urls["url_status_servico"] = nfce_group.find_value("url_status_servico") or ""
        nfce_urls["url_consulta_nfe"] = nfce_group.find_value("url_consulta_nfe") or ""
        nfce_urls["qrcode_base_url"] = nfce_group.find_value("qrcode_base_url") or ""
        nfce_urls["qrcode_check_url"] = nfce_group.find_value("qrcode_check_url") or ""

    for url in nfce_urls:
        if not nfce_urls[url]:
            logger.warning("NFCe URL missing: [{}]".format(url))

    return nfce_urls


def _get_config_group(config, group_path):
    group = config.find_group(group_path)
    configs = []
    if group:
        for group in group.groups:
            config = {group.descr: group.name}
            for key in group.keys:
                config[key.name] = key.values[0]
            configs.append(config)
    return configs


def _build_nfe_builders(
        serie,
        mod,
        ambiente,
        fiscal_parameter_controller,
        sale_line_util,
        cid_token,
        csc,
        qrcode_base_url,
        nfe_signer,
        versao_ws,
        qrcode_check_url,
        software_developer,
        ibtp_tax_processor,
        customer_text_field,
        tenders_type,
        tip_builder_common,
        store_model
):

    ide_builder = nfcebuilder.IdeBuilder(serie, mod, ambiente, versao_ws, tenders_type, store_model)
    emit_builder = nfcebuilder.EmitBuilder(store_model)

    crt = store_model.crt
    dest_builder = nfcebuilder.DestinatarioBuilder(ambiente)
    inf_resp_tec_builder = nfcebuilder.InfRespTecBuilder(software_developer)
    det_builder = nfcebuilder.DetBuilder(ambiente, crt, fiscal_parameter_controller, sale_line_util, versao_ws)
    det_tip_builder = nfcebuilder.DetTipBuilder(det_builder, tip_builder_common)
    det_list_builder = nfcebuilder.DetListBuilder(det_builder, det_tip_builder)
    total_builder = nfcebuilder.TotalBuilder(versao_ws)
    transp_builder = nfcebuilder.TranspBuilder(store_model)
    inf_intermed_builder = nfcebuilder.InfIntermedBuilder()
    pag_builder = nfcebuilder.PagBuilder(versao_ws, tenders_type)
    adic_builder = nfcebuilder.InfAdicBuilder(customer_text_field, ibtp_tax_processor)

    builders = [ide_builder,
                emit_builder,
                dest_builder,
                det_list_builder,
                total_builder,
                transp_builder,
                pag_builder,
                inf_intermed_builder,
                adic_builder,
                inf_resp_tec_builder]

    c_uf = store_model.state_code
    cnpj_contribuinte = store_model.cnpj

    inf_nfe_builder = [nfcebuilder.InfNfeBuilder(builders, c_uf, cnpj_contribuinte, mod, serie, versao_ws)]
    nfe_supl_builder = nfcebuilder.InfNFeSuplBuilder(ambiente, cid_token, csc, qrcode_base_url, qrcode_check_url, versao_ws)
    nfe_builder = nfcebuilder.NfeBuilder(inf_nfe_builder, nfe_signer, nfe_supl_builder)
    return nfe_builder


def _build_sat_builders(
    crt,                            # type: int
    cnpj_contribuinte,              # type: str
    inscr_estadual,                 # type: str
    cnpj_sw,                        # type: str
    sign_ac,                        # type: str
    fiscal_parameter_controller,    # type: FiscalParameterController
    sale_line_util,                 # type: SaleLineUtil
    tip_config,                     # type: Dict[str, Any]
    ibtp_tax_processor,             # type: IbptTaxProcessor
    customer_text_field,            # type: str
    tenders_type,                   # type: Dict[str, Any]
    tip_builder_common,             # type: TipBuilderCommon
    layout_version,                 # type: str
):
    # type: (...) -> CfeBuilder

    ide_builder = satbuilder.IdeBuilder(cnpj_sw, sign_ac)
    emit_builder = satbuilder.EmitBuilder(cnpj_contribuinte, inscr_estadual)
    dest_builder = satbuilder.DestinatarioBuilder()
    det_builder = satbuilder.DetBuilder(crt, fiscal_parameter_controller, sale_line_util, ibtp_tax_processor)
    det_tip_builder = satbuilder.DetTipBuilder(det_builder, tip_config, tip_builder_common)
    det_list_builder = satbuilder.DetListBuilder(det_builder, det_tip_builder)
    total_builder = satbuilder.TotalBuilder(ibtp_tax_processor)
    pgto_builder = satbuilder.PgtoBuilder(tenders_type)
    adic_builder = satbuilder.InfAdicBuilder(customer_text_field, ibtp_tax_processor)
    builders = [ide_builder, emit_builder, dest_builder, det_list_builder, total_builder, pgto_builder, adic_builder]
    inf_cfe_builder = [
        satbuilder.InfCfeBuilder(
            part_builders=builders,
            layout_version=layout_version,
        )
    ]
    cfe_builder = satbuilder.CfeBuilder(inf_cfe_builder)
    return cfe_builder


def _read_software_developer_informations(config, uf):
    # type: (Configuration, unicode) -> Union[SoftwareDeveloper, None]
    sv_group = config.find_group("FiscalWrapper.SoftwareDeveloper")
    software_developer = SoftwareDeveloper()

    if sv_group:
        software_developer.cnpj = str(sv_group.find_value("CNPJ") or "")
        software_developer.contact = str(sv_group.find_value("Contact") or "")
        software_developer.email = str(sv_group.find_value("Email") or "")
        software_developer.phone = str(sv_group.find_value("Phone") or "")
        software_developer.csrt = str(sv_group.find_value("CSRT") or "")
        software_developer.id_csrt = str(sv_group.find_value("idCSRT") or "")

        enabled_states = sv_group.find_values("States")
        software_developer.enabled = True if uf in enabled_states else False

    return software_developer


def _limit_max_length(value, max_length):
    return value.strip()[:max_length]


def _get_tip_config(config):
    try:
        tip_name = config.find_value("FiscalWrapper.TipConfigs.TipName") or None
        tip_code = config.find_value("FiscalWrapper.TipConfigs.TipCode")
        if tip_name is None:
            return {}
        return {'TipName': tip_name, 'TipCode': tip_code}
    except Exception as _:
        return None


def _get_fiscal_coupon_configurations(
    config,     # type: eTree.XML
):
    # type: (...) -> Dict[str, bool]

    try:
        return dict(
            operator=(config.find_value("FiscalWrapper.CouponAttrs.OPERATOR") or "True").lower() == "true",
            order_or_table=(config.find_value("FiscalWrapper.CouponAttrs.ORDER_OR_TABLE") or "True").lower() == "true",
            pos_id=(config.find_value("FiscalWrapper.CouponAttrs.POS_ID") or "True").lower() == "true",
        )
    except Exception as _:
        return {}


def _get_tip_config_by_range(config, fiscal_parameters):
    # type: (Dict, Dict) -> Union[dict, List[dict]]

    try:
        groups = _get_config_group(config, "FiscalWrapper.TipConfigs.TipCode")
        if len(groups) == 0:
            return {}

        configurations = {}
        for group in groups:
            configurations[str(group.get("id"))] = _get_configuration_tip(group, fiscal_parameters)

        return sorted(configurations.values(), key=itemgetter('PercentageUpper'))
    except Exception as _:
        return {}


def _get_configuration_tip(group, fiscal_parameters):
    # type: (Dict, Dict) -> Dict

    tip_name = str(group.get("TipName"))
    max_percentage = float(group.get("PercentageUpper") or 99999)
    tip_id = str(group.get("id"))
    key = (tip_id, "ICMS")
    exempts_code = (40, 41, 50, 60)
    exempt = True if key not in fiscal_parameters or fiscal_parameters[key] in exempts_code else False
    return {'TipName': tip_name, 'TipCode': tip_id, 'PercentageUpper': max_percentage, "exempt": exempt}


def _get_font_size_withdrawal_panel_mobile(
    config,     # type: Any
):
    # type: (...) -> str

    font_size = (config.find_value("FiscalWrapper.PrinterCustomizations.FontSizeWithdrawalPanelMobile") or "large")
    font_size = font_size.lower()

    if font_size in ['small', 'large']:
        return font_size

    return "large"


def _get_danfe_font_size_mobile(
  config,   # type: Any
):
    # type: (...) -> str

    font_size = (config.find_value("FiscalWrapper.PrinterCustomizations.DanfePOSMobileFontSize") or "small")
    return font_size.lower()


def _get_order_ref_ids(config):
    default_ref_ids = "PAGER_NUMBER|CUSTOMER_NAME|ORDER_ID"
    return config.find_value("FiscalWrapper.OrderReferenceIdentifier", default_ref_ids).upper().split("|")


def _get_development_serie_and_order_id_increment():
    if os.environ.get("DEBUG", "").lower() == "true":
        initial_order_id_inc_file_path = os.path.join(os.environ["BUNDLEDIR"], "initial_order_id.txt")
        if os.path.exists(initial_order_id_inc_file_path):
            with open(initial_order_id_inc_file_path, "rb") as f:
                file_content = f.read().decode("utf-8").strip()
                serie_inc, initial_order_id_inc = file_content.split(";")
                return int(serie_inc), int(initial_order_id_inc)

    return None, None


def _get_loyalty_configurations(
    mb_context,     # type: MBEasyContext
    config,         # type: eTree.XML
):
    # type: (...) -> LoyaltyConfigurations

    try:
        print_loyalty = config.find_value("FiscalWrapper.Loyalty.PrintLoyaltyCouponAfterFiscalization") or "False"
        text_to_print = config.find_value("FiscalWrapper.Loyalty.TextToPrintInLoyaltyCoupon") or ""
        return LoyaltyConfigurations(
            print_loyalty_coupon_after_fiscalization=print_loyalty.lower() == "true",
            text_to_print_in_loyalty_coupon=text_to_print,
            store_name=read_swconfig(
                mbcontext=mb_context,
                key="Store.Name",
            ) or "",
        )
    except Exception as _:
        return LoyaltyConfigurations(
            print_loyalty_coupon_after_fiscalization=False,
            text_to_print_in_loyalty_coupon="",
            store_name="",
        )


if __name__ == "__main__":
    main()



import json

from msgbus import MBEasyContext, TK_SYS_ACK
from pos_model import TaxItem, SaleItem
from typing import Dict, List


class CalculateTaxesItemCommon(object):
    def __init__(self, mb_context):
        # type: (MBEasyContext) -> None

        self.mb_context = mb_context

    def get_taxes_item(
        self,               # type: CalculateTaxesItemCommon
        sale_line,          # type: SaleItem
        taxes_calc,         # type: List
        base_reduction,     # type: Dict
        cst_icms,           # type: Dict
    ):
        # type: (...) -> List[TaxItem]

        data = json.dumps(dict(
            partCode=sale_line.part_code,
            productName=sale_line.product_name,
            quantity=sale_line.quantity,
            unitPrice=sale_line.unit_price,
            itemPrice=sale_line.item_price,
            taxesCalc=taxes_calc,
            baseReduction=base_reduction,
            cst=cst_icms,
        )).encode('utf-8')

        msg = self.mb_context.MB_EasyEvtSend("TAX_CALC_ITEM", "", data, synchronous=True)
        if msg.token == TK_SYS_ACK:
            return json.loads(msg.data, object_hook=self.tax_decode)

        return []

    @staticmethod
    def tax_decode(data):
        # type: (Dict) -> TaxItem

        tax = TaxItem()
        tax.tax_code = data.get("code")
        tax.tax_name = data.get("tax_name")
        tax.tax_rate = data.get("tax_rate")
        tax.tax_index = data.get("tax_index")
        tax.tax_rule_id = data.get("tax_rule_id")
        tax.base_amount_ad = data.get("base_amount_ad")
        tax.base_amount_bd = data.get("base_amount_bd")
        tax.base_reduced_amount = data.get("base_reduced_amount")
        tax.tax_amount_ad = data.get("tax_amount_ad")
        tax.tax_amount_bd = data.get("tax_amount_bd")
        tax.tax_included = data.get("tax_included")

        return tax



from typing import List


class BumpBarCommand(object):
    def __init__(self, name, keys):
        # type: (unicode, List[unicode]) -> BumpBarCommand
        self.name = name
        self.keys = keys



# gstin.py - functions for handling Indian VAT numbers
#
# Copyright (C) 2021 Gaurav Chauhan
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""GSTIN (Goods and Services Tax identification number, Indian VAT number).

The Goods and Services Tax identification number (GSTIN) is a 15 digit unique
identifier assigned to all business entities in India registered under the
Goods and Services Tax (GST) Act, 2017.

Each GSTIN begins with a 2 digit state code, the next 10 characters are the
holder's PAN, the 13th character is an alphanumeric digit that represents the
number of GSTIN registrations made in a state or union territory for same the
PAN, the 14th character is 'Z' and the last character is an alphanumeric
check digit calculated using Luhn mod 36 algorithm.

More information:

* https://bajajfinserv.in/insights/what-is-goods-and-service-tax-identification-number
* https://ddvat.gov.in/docs/List%20of%20State%20Code.pdf
* https://en.wikipedia.org/wiki/Goods_and_Services_Tax_(India)

>>> validate('27AAPFU0939F1ZV')
'27AAPFU0939F1ZV'
>>> validate('27AAPFU0939F1Z')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('369296450896540')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('27AAPFU0939F1AA')
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('27AAPFU0939F1ZO')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> to_pan('27AAPFU0939F1ZV')
'AAPFU0939F'
>>> info('27AAPFU0939F1ZV')['state']
'Maharashtra'
"""

import re

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.in_ import pan
from stdnum.util import clean


_GSTIN_RE = re.compile(r'^[0-9]{2}[A-Z]{5}[0-9]{4}[A-Z][0-9A-Z]{3}$')

_ALPHABET = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'

_STATE_CODES = {
    '01': 'Jammu and Kashmir',
    '02': 'Himachal Pradesh',
    '03': 'Punjab',
    '04': 'Chandigarh',
    '05': 'Uttarakhand',
    '06': 'Haryana',
    '07': 'Delhi',
    '08': 'Rajasthan',
    '09': 'Uttar Pradesh',
    '10': 'Bihar',
    '11': 'Sikkim',
    '12': 'Arunachal Pradesh',
    '13': 'Nagaland',
    '14': 'Manipur',
    '15': 'Mizoram',
    '16': 'Tripura',
    '17': 'Meghalaya',
    '18': 'Assam',
    '19': 'West Bengal',
    '20': 'Jharkhand',
    '21': 'Orissa',
    '22': 'Chattisgarh',
    '23': 'Madhya Pradesh',
    '24': 'Gujarat',
    '25': 'Daman and Diu',
    '26': 'Dadar and Nagar Haveli',
    '27': 'Maharashtra',
    '28': 'Andhra Pradesh',
    '29': 'Karnataka',
    '30': 'Goa',
    '31': 'Lakshadweep',
    '32': 'Kerala',
    '33': 'Tamil Nadu',
    '34': 'Puducherry',
    '35': 'Anadaman and Nicobar Islands',
    '36': 'Telangana',
    '37': 'Andhra Pradesh (New)',
}


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').upper().strip()


def validate(number):
    """Check if the number provided is a valid GSTIN. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if len(number) != 15:
        raise InvalidLength()
    if not _GSTIN_RE.match(number):
        raise InvalidFormat()
    if number[:2] not in _STATE_CODES or number[12] == '0' or number[13] != 'Z':
        raise InvalidComponent()
    pan.validate(number[2:12])
    luhn.validate(number, _ALPHABET)
    return number


def is_valid(number):
    """Check if the number provided is a valid GSTIN. This checks the length,
    formatting and check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def to_pan(number):
    """Convert the number to a PAN."""
    number = compact(number)
    return number[2:12]


def info(number):
    """Provide information that can be decoded locally from GSTIN (without
    API)."""
    number = validate(number)
    return {
        'state': _STATE_CODES.get(number[:2]),
        'pan': number[2:12],
        'holder_type': pan.info(number[2:12])['holder_type'],
        'initial': number[6],
        'registration_count': _ALPHABET.index(number[12]),
    }
